library IEEE;
library work;

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.multiboot_pkg.ALL;

--! Used to import ICAP primitive
library UNISIM;
use UNISIM.VComponents.all;

entity multiboot_core is
   port(clk         : in  STD_LOGIC;
        rst         : in  STD_LOGIC;

        CTR0_i      : in  r_CTR0;
        CTR1_o      : out r_CTR1;
        STAT_o      : out r_STAT;
        MBA_i       : in  r_BAR;
        GBA_i       : in  r_BAR;
        MBA_ICAP_o  : out STD_LOGIC_VECTOR (23 downto 0);
        GBA_ICAP_o  : out STD_LOGIC_VECTOR (23 downto 0));
end multiboot_core;

architecture Behavioral of multiboot_core is

   type ICAP_fsm is (R0_RESET,
                     S0_IDLE,
                     S1A_MULTIBOOT,
                     S1B_READBACK);

   --! The fsm for multiboot can be found in ug380 pg126
   --! Added four noop states as it is done in example xt059.pdf
   type multi_ICAP_fsm is (R0_RESET,
                           S0_IDLE,
                           S1_DUMMY,
                           S2_SYNC_H,
                           S3_SYNC_L,
                           S4_GEN1_CMD,
                           S5_GEN1_ADDR,
                           S6_GEN2_CMD,
                           S7_GEN2_ADDR,
                           S8_GEN3_CMD,
                           S9_GEN3_ADDR,
                           S10_GEN4_CMD,
                           S11_GEN4_ADDR,
                           S12_CMD_CMD,
                           S13_IPROG,
                           S14_NOOP,
                           S15_NOOP,
                           S16_NOOP,
                           S17_NOOP,
                           Q0_OK);

   --! The fsm for readbacks can be found in ug380 pg109
   type readback_ICAP_fsm is (R0_RESET,
                              S0_IDLE,
                              S1_DUMMY1, S2_DUMMY2,
                              S3_SYNC_H,
                              S4_SYNC_L,
                              S5_NOOP,
                              S6A_READ_CMD,
                              S7A_NOOP, S8A_NOOP, S9A_NOOP, S10A_NOOP,
                              --! The fetch reg state will take several cycles
                              S11A_FETCH_REG,
                              S6B_READ_CMD,
                              S7B_NOOP, S8B_NOOP, S9B_NOOP, S10B_NOOP,
                              --! The fetch reg state will take several cycles
                              S11B_FETCH_REG,
                              S12_WRITE_CMD,
                              S13_DESYNC_CMD,
                              S14_NOOP, S15_NOOP,
                              Q0_OK);

   signal s_ICAP_fsm         : ICAP_fsm;
   signal s_ICAP_fsm_d0      : ICAP_fsm;
   signal s_multiboot_fsm    : multi_ICAP_fsm;
   signal s_multiboot_fsm_d0 : multi_ICAP_fsm;
   signal s_readback_fsm     : readback_ICAP_fsm;
   signal s_readback_fsm_d0  : readback_ICAP_fsm;

   signal s_icap_busy        : STD_LOGIC;
   signal s_icap_o           : STD_LOGIC_VECTOR(15 downto 0);
   signal s_icap_ce_n        : STD_LOGIC := '1'; --! Enable in ICAP is active low
   signal s_icap_i           : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
   signal s_icap_tmp         : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
   signal s_icap_wr_n        : STD_LOGIC := '1';

   signal s_CTR0             : r_CTR0;
   --! This is a copy of the instruction to be performed, done in S0_IDLE
   signal s_CTR0_core        : r_CTR0;
   signal s_CTR1             : r_CTR1;
   signal s_STAT_ICAP        : r_STAT := c_STAT_default;
   signal s_MBA              : r_BAR;
   signal s_GBA              : r_BAR;
   --! It should be noted that xBAs go to 1 to let the inconsistency be
   --! detected.
   signal s_MBA_ICAP         : STD_LOGIC_VECTOR(23 downto 0) := (others => '1');
   signal s_GBA_ICAP         : STD_LOGIC_VECTOR(23 downto 0) := (others => '1');

   signal s_rd_ICAP_regA     : STD_LOGIC_VECTOR(15 downto 0) := NULL_value;
   signal s_rd_ICAP_regB     : STD_LOGIC_VECTOR(15 downto 0) := NULL_value;
   signal s_wr_ICAP_gen1     : STD_LOGIC_VECTOR(15 downto 0) := NULL_value;
   signal s_wr_ICAP_gen2     : STD_LOGIC_VECTOR(15 downto 0) := NULL_value;
   signal s_wr_ICAP_gen3     : STD_LOGIC_VECTOR(15 downto 0) := NULL_value;
   signal s_wr_ICAP_gen4     : STD_LOGIC_VECTOR(15 downto 0) := NULL_value;

begin

   s_CTR0            <= CTR0_i;
   CTR1_o            <= s_CTR1;
   STAT_o            <= s_STAT_ICAP;
   s_MBA             <= MBA_i;
   s_GBA             <= GBA_i;
   MBA_ICAP_o        <= s_MBA_ICAP;
   GBA_ICAP_o        <= s_GBA_ICAP;
   order_ICAP(s_icap_tmp, s_icap_i);

   ICAP_SPARTAN6_inst : ICAP_SPARTAN6
   generic map (-- Pre-programmed 'Device ID code'
                DEVICE_ID         => to_bitvector(c_DEVICE_ID_XC6SLX45T),
                -- No simulation possible of ICAP
                SIM_CFG_FILE_NAME => "../rtl/basic_trigger_top.rbt")
   port map (BUSY   => s_icap_busy,  -- 1-bit  output Busy/Ready output
             O      => s_icap_o,     -- 16-bit output Configuartion data output bus
             CE     => s_icap_ce_n,  -- 1-bit  input  Active-Low ICAP Enable input
             CLK    => clk,          -- 1-bit  input  Clock input
             I      => s_icap_i,     -- 16-bit input  Configuration data input bus
             WRITE  => s_icap_wr_n); -- 1-bit  input  Read/Write control input


   --! @brief                  Process to update CTR1
   --! @param  s_readback_fsm  subfsm of the readback
   p_xBAC_ICAP: process (s_MBA_ICAP,
                         s_GBA_ICAP,
                         s_wr_ICAP_gen1,
                         s_wr_ICAP_gen2,
                         s_wr_ICAP_gen3,
                         s_wr_ICAP_gen4)
      variable v_MBA_wr  : STD_LOGIC_VECTOR(23 downto 0);
      variable v_GBA_wr  : STD_LOGIC_VECTOR(23 downto 0);
   begin
      v_MBA_wr :=   s_wr_ICAP_gen1(7  downto 0)
                  & s_wr_ICAP_gen2;
      v_GBA_wr :=   s_wr_ICAP_gen3(7  downto 0)
                  & s_wr_ICAP_gen4;

      s_CTR1.MBAC <= '0';
      s_CTR1.GBAC <= '0';

      if s_MBA_ICAP = v_MBA_wr then
         s_CTR1.MBAC <= '1';
      end if;
      if s_GBA_ICAP = v_GBA_wr then
         s_CTR1.GBAC <= '1';
      end if;
   end process;

   --! @brief                  Process to update the FSM_RUN bit for lock
   --!                         control
   --! @param s_ICAP_fsm       Main fsm for ICAP
   --! @param s_multiboot_fsm  subfsm for writes into ICAP
   --! @param s_readback_fsm   subfsm for reads into ICAP
   p_CTR1_FSM_RUN : process (s_ICAP_fsm,
                             s_multiboot_fsm,
                             s_readback_fsm)
   begin
      s_CTR1.FSM_RUN <= '0';
      case s_ICAP_fsm is
         when R0_RESET      =>
            s_CTR1.FSM_RUN <= c_CTR1_default.FSM_RUN;
         when S1A_MULTIBOOT =>
            s_CTR1.FSM_RUN <= '1';
            if s_multiboot_fsm = Q0_OK then
               s_CTR1.FSM_RUN <= '0';
            end if;
         when S1B_READBACK  =>
            s_CTR1.FSM_RUN <= '1';
            if s_readback_fsm = Q0_OK then
               s_CTR1.FSM_RUN <= '0';
            end if;
         when others =>
      end case;
   end process p_CTR1_FSM_RUN;

   --! @brief      Process to update CTR1
   --! @param clk  Main clock
   p_CTR1: process (clk)
   begin
      if rising_edge(clk) then
         s_ICAP_fsm_d0      <= s_ICAP_fsm;
         s_multiboot_fsm_d0 <= s_multiboot_fsm;
         s_readback_fsm_d0  <= s_readback_fsm;
         s_CTR1.OP_ACK      <= c_CTR1_default.OP_ACK;
         s_CTR1.BAD_OP      <= c_CTR1_default.BAD_OP;
         s_CTR1.x           <= c_CTR1_default.x;
         --! Watchout! CTR1 is one-clock delayed
         case s_ICAP_fsm is
            when R0_RESET      =>
               --! CTR1.OP_ACK from an ICAP write (multiboot_fsm)
               if s_ICAP_fsm_d0 = S1A_MULTIBOOT then
                  case s_multiboot_fsm_d0 is
                     when Q0_OK  =>
                        s_CTR1.OP_ACK <= '1';
                     when others =>
                        s_CTR1.BAD_OP <= '1';
                  end case;
               end if;
               --! CTR1.OP_ACK from an ICAP readback
               if s_ICAP_fsm_d0 = S1B_READBACK  then
                  case s_readback_fsm_d0 is
                     when Q0_OK  =>
                        s_CTR1.OP_ACK <= '1';
                     when others =>
                        s_CTR1.BAD_OP <= '1';
                  end case;
               end if;
            when others        =>
               null;
         end case;
      end if;
   end process;

   --! @brief                  Process to set the rd/wr bit
   --! @param  s_ICAP_fsm      fsm of the ICAP
   --! @param  s_multiboot_fsm subfsm of the multiboot
   --! @param  s_readback_fsm  subfsm of the readback
   p_ICAP_rd_wr : process (s_ICAP_fsm,
                           s_multiboot_fsm,
                           s_readback_fsm)
   begin
      s_icap_wr_n <= '1';
      case s_ICAP_fsm is
         when S1A_MULTIBOOT =>
            s_icap_wr_n <= '0';
         when S1B_READBACK  =>
            s_icap_wr_n <= '0';
               case s_readback_fsm is
                 when S11A_FETCH_REG =>
                    s_icap_wr_n <= '1';
                 when S11B_FETCH_REG =>
                    s_icap_wr_n <= '1';
                 when others       =>
                    null;
               end case;
         when others        =>
            null;
      end case;
   end process p_ICAP_rd_wr;

   --! @brief                  Enable process for the ICAP primitive
   --! @param  s_ICAP_fsm      fsm of the ICAP
   --! @param  s_multiboot_fsm subfsm of the multiboot
   --! @param  s_readback_fsm  subfsm of the readback
   p_ICAP_en : process(s_ICAP_fsm,
                       s_multiboot_fsm,
                       s_readback_fsm)
   begin
      s_icap_ce_n <= '1';
      case s_ICAP_fsm is
         when S1A_MULTIBOOT =>
            s_icap_ce_n <= '0';
         when S1B_READBACK  =>
            s_icap_ce_n <= '0';
         when others        =>
            null;
         end case;
   end process p_ICAP_en;

   --! @brief       xBA_ICAP port update upon ICAP primitive fsm
   --! @param  clk  Main clock
   p_ICAP_read_port: process(clk)
      variable v_STAT  : STD_LOGIC_VECTOR(r_STAT'a_length - 1 downto 0);
   begin
      if rising_edge(clk) then
         case s_readback_fsm is
            when S11A_FETCH_REG =>
               case s_CTR0.OP is
                  when OP_RD_STAT =>
                     v_STAT      := s_icap_o;
                     s_STAT_ICAP <= f_STAT(v_STAT);
                  when OP_RD_MBA  =>
                     s_MBA_ICAP (15 downto  0) <= s_icap_o;
                  when OP_RD_GBA  =>
                     s_GBA_ICAP (15 downto  0) <= s_icap_o;
                  when others     =>
                     null;
               end case;
            when S11B_FETCH_REG =>
               case s_CTR0.OP is
                  when OP_RD_MBA  =>
                     s_MBA_ICAP (23 downto 16) <= s_icap_o(7 downto 0);
                  when OP_RD_GBA  =>
                     s_GBA_ICAP (23 downto 16) <= s_icap_o(7 downto 0);
                  when others     =>
                     null;
               end case;
            when others         =>
               null;
         end case;
      end if;
   end process;

   --! @brief                  Input port update process for the ICAP primitive
   --! @param  s_multiboot_fsm subfsm of the multiboot
   --! @param  s_readback_fsm  subfsm of the readback
   p_ICAP_write_port: process(s_ICAP_fsm,
                              s_multiboot_fsm,
                              s_readback_fsm)
   begin
      s_icap_tmp <= NULL_value;
      case s_ICAP_fsm is
         when S1A_MULTIBOOT =>
            case s_multiboot_fsm is
               when R0_RESET      =>
                  null;
               when S0_IDLE       =>
                  null;
               when S1_DUMMY      =>
                  s_icap_tmp <= DUMMY;
               when S2_SYNC_H     =>
                  s_icap_tmp <= SYNC_H;
               when S3_SYNC_L     =>
                  s_icap_tmp <= SYNC_L;
               when S4_GEN1_CMD   =>
                  s_icap_tmp <= CMD_wr_GEN1;
               when S5_GEN1_ADDR  =>
                  s_icap_tmp <= s_wr_ICAP_gen1;
               when S6_GEN2_CMD   =>
                  s_icap_tmp <= CMD_wr_GEN2;
               when S7_GEN2_ADDR  =>
                  s_icap_tmp <= s_wr_ICAP_gen2;
               when S8_GEN3_CMD   =>
                  s_icap_tmp <= CMD_wr_GEN3;
               when S9_GEN3_ADDR  =>
                  s_icap_tmp <= s_wr_ICAP_gen3;
               when S10_GEN4_CMD  =>
                  s_icap_tmp <= CMD_wr_GEN4;
               when S11_GEN4_ADDR =>
                  s_icap_tmp <= s_wr_ICAP_gen4;
               when S12_CMD_CMD   =>
                  s_icap_tmp <= CMD_wr_CMD;
               when S13_IPROG     =>
                  s_icap_tmp <= CMD_IPROG;
               when S14_NOOP      =>
                  s_icap_tmp <= NOOP;
               when S15_NOOP      =>
                  s_icap_tmp <= NOOP;
               when S16_NOOP      =>
                  s_icap_tmp <= NOOP;
               when S17_NOOP      =>
                  s_icap_tmp <= NOOP;
               when Q0_OK         =>
                  null;
               when others        =>
                  null;
            end case;
         when S1B_READBACK  =>
            case s_readback_fsm is
               when R0_RESET    =>
                  null;
               when S1_DUMMY1      =>
                  s_icap_tmp <= DUMMY;
               when S2_DUMMY2      =>
                  s_icap_tmp <= DUMMY;
               when S3_SYNC_H      =>
                  s_icap_tmp <= SYNC_H;
               when S4_SYNC_L      =>
                  s_icap_tmp <= SYNC_L;
               when S5_NOOP        =>
                  s_icap_tmp <= NOOP;
               when S6A_READ_CMD    =>
                  s_icap_tmp <= s_rd_ICAP_regA;
               when S7A_NOOP        =>
                  s_icap_tmp <= NOOP;
               when S8A_NOOP        =>
                  s_icap_tmp <= NOOP;
               when S9A_NOOP        =>
                  s_icap_tmp <= NOOP;
               when S10A_NOOP       =>
                  s_icap_tmp <= NOOP;
               when S11A_FETCH_REG  =>
                  null;
               when S6B_READ_CMD    =>
                  s_icap_tmp <= s_rd_ICAP_regB;
               when S7B_NOOP        =>
                  s_icap_tmp <= NOOP;
               when S8B_NOOP        =>
                  s_icap_tmp <= NOOP;
               when S9B_NOOP        =>
                  s_icap_tmp <= NOOP;
               when S10B_NOOP       =>
                  s_icap_tmp <= NOOP;
               when S11B_FETCH_REG  =>
                  null;
               when S12_WRITE_CMD  =>
                  s_icap_tmp <= CMD_wr_CMD;
               when S13_DESYNC_CMD =>
                  s_icap_tmp <= DESYNC;
               when S14_NOOP       =>
                  s_icap_tmp <= NOOP;
               when S15_NOOP       =>
                  s_icap_tmp <= NOOP;
               when Q0_OK          =>
                  null;
               when others         =>
                  null;
            end case;
         when others  =>
            null;
      end case;
   end process p_ICAP_write_port;


   --! @brief      readback subfsm dependant on the value of the ICAP_fsm
   --! @param clk  Main clock
   p_readback_fsm: process(clk)

      procedure update_readback_fsm is
      begin
         s_readback_fsm <= R0_RESET;
         case s_readback_fsm is
            when R0_RESET       =>
               s_readback_fsm <= S0_IDLE;
            when S0_IDLE        =>
               s_readback_fsm <= S1_DUMMY1;
            when S1_DUMMY1      =>
               s_readback_fsm <= S2_DUMMY2;
            when S2_DUMMY2      =>
               s_readback_fsm <= S3_SYNC_H;
            when S3_SYNC_H      =>
               s_readback_fsm <= S4_SYNC_L;
            when S4_SYNC_L      =>
               s_readback_fsm <= S5_NOOP;
            when S5_NOOP        =>
               s_readback_fsm <= S6A_READ_CMD;
            when S6A_READ_CMD    =>
               s_readback_fsm <= S7A_NOOP;
            when S7A_NOOP        =>
               s_readback_fsm <= S8A_NOOP;
            when S8A_NOOP        =>
               s_readback_fsm <= S9A_NOOP;
            when S9A_NOOP        =>
               s_readback_fsm <= S10A_NOOP;
            when S10A_NOOP       =>
               s_readback_fsm <= S11A_FETCH_REG;
            when S11A_FETCH_REG  =>
               if s_rd_ICAP_regA = STD_LOGIC_VECTOR(CMD_rd_STAT) then
                  s_readback_fsm <= S12_WRITE_CMD;
               else
                  s_readback_fsm <= S6B_READ_CMD;
               end if;
            when S6B_READ_CMD    =>
               s_readback_fsm <= S7B_NOOP;
            when S7B_NOOP        =>
               s_readback_fsm <= S8B_NOOP;
            when S8B_NOOP        =>
               s_readback_fsm <= S9B_NOOP;
            when S9B_NOOP        =>
               s_readback_fsm <= S10B_NOOP;
            when S10B_NOOP       =>
               s_readback_fsm <= S11B_FETCH_REG;
            when S11B_FETCH_REG  =>
               s_readback_fsm <= S12_WRITE_CMD;
            when S12_WRITE_CMD  =>
               s_readback_fsm <= S13_DESYNC_CMD;
            when S13_DESYNC_CMD =>
               s_readback_fsm <= S14_NOOP;
            when S14_NOOP       =>
               s_readback_fsm <= S15_NOOP;
            when S15_NOOP       =>
               s_readback_fsm <= Q0_OK;
            when Q0_OK          =>
               s_readback_fsm <= S0_IDLE;
            when others         =>
               null;
         end case;
      end procedure;

   begin
      if rising_edge(clk) then
         s_readback_fsm <= R0_RESET;
         case s_ICAP_fsm is
            when S1B_READBACK   =>
               update_readback_fsm;
               case s_CTR0_core.OP is
                  when OP_RD_MBA     =>
                     s_rd_ICAP_regA <= CMD_rd_GEN1;
                     s_rd_ICAP_regB <= CMD_rd_GEN2;
                  when OP_RD_GBA     =>
                     s_rd_ICAP_regA <= CMD_rd_GEN3;
                     s_rd_ICAP_regB <= CMD_rd_GEN4;
                  when OP_RD_STAT    =>
                     s_rd_ICAP_regA <= CMD_rd_STAT;
                  when others        =>
               end case;
            when others =>
               null;
         end case;
      end if;
   end process p_readback_fsm;


   --! @brief      readback subfsm dependant on the value of the ICAP_fsm
   --! @param clk  Main clock
   p_multiboot_fsm: process(clk)

      procedure update_multiboot_fsm(begin_fsm  : multi_ICAP_fsm;
                                     end_fsm    : multi_ICAP_fsm) is
      begin
         s_multiboot_fsm <=  R0_RESET;
         case s_multiboot_fsm is
            when R0_RESET        =>
               s_multiboot_fsm <=  S0_IDLE;
            when S0_IDLE         =>
               s_multiboot_fsm <=  S1_DUMMY;
            when S1_DUMMY        =>
               s_multiboot_fsm <=  S2_SYNC_H;
            when S2_SYNC_H       =>
               s_multiboot_fsm <=  S3_SYNC_L;
            when S3_SYNC_L       =>
               s_multiboot_fsm <= R0_RESET;
               case begin_fsm is
                  when S4_GEN1_CMD  =>
                     s_multiboot_fsm <= S4_GEN1_CMD;
                  when S8_GEN3_CMD  =>
                     s_multiboot_fsm <= S8_GEN3_CMD;
                  when S12_CMD_CMD  =>
                     s_multiboot_fsm <= S12_CMD_CMD;
                  when others =>
                     null;
               end case;
            when S4_GEN1_CMD     =>
               s_multiboot_fsm <=  S5_GEN1_ADDR;
            when S5_GEN1_ADDR    =>
               s_multiboot_fsm <=  S6_GEN2_CMD;
            when S6_GEN2_CMD     =>
               s_multiboot_fsm <=  S7_GEN2_ADDR;
            when S7_GEN2_ADDR    =>
               s_multiboot_fsm <=  S8_GEN3_CMD;
               case end_fsm is
                  when Q0_OK        =>
                     null;
                  when S7_GEN2_ADDR    =>
                     s_multiboot_fsm <= S12_CMD_CMD;
                  when others =>
                     s_multiboot_fsm <= R0_RESET;
               end case;
            when S8_GEN3_CMD     =>
               s_multiboot_fsm <=  S9_GEN3_ADDR;
            when S9_GEN3_ADDR    =>
               s_multiboot_fsm <=  S10_GEN4_CMD;
            when S10_GEN4_CMD    =>
               s_multiboot_fsm <=  S11_GEN4_ADDR;
            when S11_GEN4_ADDR   =>
               s_multiboot_fsm <=  S12_CMD_CMD;
               case end_fsm is
                  when Q0_OK        =>
                     null;
                  when S11_GEN4_ADDR   =>
                     null;
                  when others =>
                     s_multiboot_fsm <= R0_RESET;
               end case;
            when S12_CMD_CMD     =>
               s_multiboot_fsm <=  S13_IPROG;
            when S13_IPROG       =>
               s_multiboot_fsm <=  S14_NOOP;
            when S14_NOOP        =>
               s_multiboot_fsm <=  S15_NOOP;
            when S15_NOOP        =>
               s_multiboot_fsm <=  S16_NOOP;
            when S16_NOOP        =>
               s_multiboot_fsm <=  S17_NOOP;
            when S17_NOOP        =>
               s_multiboot_fsm <=  Q0_OK;
            when Q0_OK           =>
               s_multiboot_fsm <=  R0_RESET;
            when others =>
               null;
         end case;
      end procedure;

      procedure update_wr_general(position : NATURAL) is
         variable v_begin  : multi_ICAP_fsm;
         variable v_end    : multi_ICAP_fsm;
         variable v_pos    : UNSIGNED (3 downto 0);
      begin
         v_begin  := R0_RESET;
         v_end    := R0_RESET;
         v_pos    := to_unsigned(position, 4);
         case v_pos is
            when OP_FULLMULTIBOOT =>
               v_begin  := S4_GEN1_CMD;
               v_end    := Q0_OK;
            when OP_WR_MBA =>
               v_begin  := S4_GEN1_CMD;
               v_end    := S7_GEN2_ADDR;
            when OP_WR_GBA =>
               v_begin  := S8_GEN3_CMD;
               v_end    := S11_GEN4_ADDR;
            when OP_CMD_IPROG =>
               v_begin  := S12_CMD_CMD;
               v_end    := Q0_OK;
            when others =>
               null;
         end case;
         update_multiboot_fsm(v_begin, v_end);
      end procedure;

      variable v_MBA_slv  : STD_LOGIC_VECTOR(31 downto 0);
      variable v_GBA_slv  : STD_LOGIC_VECTOR(31 downto 0);
      variable v_GENERAL1  : STD_LOGIC_VECTOR(15 downto 0);
      variable v_GENERAL2  : STD_LOGIC_VECTOR(15 downto 0);
      variable v_GENERAL3  : STD_LOGIC_VECTOR(15 downto 0);
      variable v_GENERAL4  : STD_LOGIC_VECTOR(15 downto 0);

   begin
      if rising_edge(clk) then
         s_multiboot_fsm <= R0_RESET;
         v_MBA_slv := f_STD_LOGIC_VECTOR(s_MBA);
         v_GBA_slv := f_STD_LOGIC_VECTOR(s_GBA);
         case s_ICAP_fsm is
            when S1A_MULTIBOOT  =>
               case s_CTR0_core.OP is
                  when OP_FULLMULTIBOOT =>
                     update_wr_general(to_integer(s_CTR0_core.OP));
                     s_wr_ICAP_gen1 <= v_MBA_slv(15 downto  0);
                     s_wr_ICAP_gen2 <= v_MBA_slv(31 downto 16);
                     s_wr_ICAP_gen3 <= v_GBA_slv(15 downto  0);
                     s_wr_ICAP_gen4 <= v_GBA_slv(31 downto 16);
                  when OP_WR_MBA   =>
                     update_wr_general(to_integer(s_CTR0_core.OP));
                     v_MBA_slv      := f_STD_LOGIC_VECTOR(s_MBA);
                     s_wr_ICAP_gen1 <= v_MBA_slv(15 downto  0);
                     s_wr_ICAP_gen2 <= v_MBA_slv(31 downto 16);
                  when OP_WR_GBA   =>
                     update_wr_general(to_integer(s_CTR0_core.OP));
                     v_GBA_slv      := f_STD_LOGIC_VECTOR(s_GBA);
                     s_wr_ICAP_gen3 <= v_GBA_slv(15 downto  0);
                     s_wr_ICAP_gen4 <= v_GBA_slv(31 downto 16);
                  when OP_CMD_IPROG =>
                     update_wr_general(to_integer(s_CTR0_core.OP));
                  when others =>
                     null;
               end case;
            when others =>
               null;
         end case;
      end if;
   end process p_multiboot_fsm;


   --! @brief      Coarse ICAP fsm
   --! @param clk  Main clock
   p_ICAP_fsm: process(clk)
   begin
      if rising_edge(clk) then
         if rst = '1' then
            s_ICAP_fsm    <= R0_RESET;
            s_CTR0_core   <= c_CTR0_default;
         else
            case s_ICAP_fsm is
               when R0_RESET  =>
                  s_ICAP_fsm <= S0_IDLE;
               when S0_IDLE   =>
                  case s_CTR0.PEN is
                     when '0'    =>
                        s_CTR0_core   <= c_CTR0_default;
                     when others =>
                        if s_CTR1.FSM_RUN = '0'   then
                           s_CTR0_core   <= s_CTR0;
                           --! In that case we start an operation
                           if    s_CTR0.OP = OP_FULLMULTIBOOT
                              or s_CTR0.OP = OP_WR_MBA
                              or s_CTR0.OP = OP_WR_GBA
                              or s_CTR0.OP = OP_CMD_IPROG    then
                              s_ICAP_fsm <= S1A_MULTIBOOT;
                           end if;
                           if    s_CTR0.OP = OP_RD_MBA
                              or s_CTR0.OP = OP_RD_GBA
                              or s_CTR0.OP = OP_RD_STAT      then
                              s_ICAP_fsm <= S1B_READBACK;
                           end if;
                        end if;
                  end case;
               when S1A_MULTIBOOT =>
                  --! Or we attend that the s_CTR1.FSM_RUN has been
                  --! deasserted.
                  if s_multiboot_fsm = Q0_OK then
                     s_ICAP_fsm <= R0_RESET;
                  end if;
               when S1B_READBACK  =>
                  --! Or we attend that the s_CTR1.FSM_RUN has been
                  --! deasserted.
                  if s_readback_fsm  = Q0_OK   then
                     s_ICAP_fsm <= R0_RESET;
                  end if;
               when others =>
                  null;
            end case;
         end if;
      end if;
   end process p_ICAP_fsm;

end Behavioral;
