--------------------------------------------------------------------------------
-- CERN (BE-CO-HT)
-- Top level entity of CONV-TTL-BLO V1
-- http://www.ohwr.org/projects/conv-trig-blo
--------------------------------------------------------------------------------
--
-- unit name: conv_ttl_blo_v2.vhd
--
-- author: Theodor-Adrian Stana (t.stana@cern.ch)
--
-- version: 1.0
--
-- description: Top entity of CONV-TTL-BLO V1
--
-- dependencies:
--
--------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
--------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--------------------------------------------------------------------------------
library IEEE;
library unisim;

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use UNISIM.VCOMPONENTS.ALL;

use work.bicolor_led_ctrl_pkg.all;
use work.genram_pkg.all;
use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.wr_fabric_pkg.all;
use work.endpoint_pkg.all;

entity conv_ttl_blo_v2 is
  generic
  (
    g_nr_ttl_chan : natural := 6;
    g_nr_inv_chan : natural := 4
  );
  port 
  (
    -- Clock inputs
    clk20_vcxo_i           : in  std_logic;
    fpga_clk_p_i           : in  std_logic; --Using the 125MHz clock
    fpga_clk_n_i           : in  std_logic;
    
    -- LEDs
    led_ctrl0_o            : out std_logic;
    led_ctrl0_oen_o        : out std_logic;
    led_ctrl1_o            : out std_logic;
    led_ctrl1_oen_o        : out std_logic;
    led_multicast_2_0_o    : out std_logic;
    led_multicast_3_1_o    : out std_logic;
    led_wr_gmt_ttl_ttln_o  : out std_logic;
    led_wr_link_syserror_o : out std_logic;
    led_wr_ok_syspw_o      : out std_logic;
    led_wr_ownaddr_i2c_o   : out std_logic;
    
    -- I/Os for pulses
    pulse_front_led_n_o    : out std_logic_vector(g_nr_ttl_chan downto 1);
    pulse_rear_led_n_o     : out std_logic_vector(g_nr_ttl_chan downto 1);
    fpga_input_ttl_n_i     : in  std_logic_vector(g_nr_ttl_chan downto 1);
    fpga_out_ttl_o         : out std_logic_vector(g_nr_ttl_chan downto 1);
    fpga_blo_in_i          : in  std_logic_vector(g_nr_ttl_chan downto 1);
    fpga_trig_blo_o        : out std_logic_vector(g_nr_ttl_chan downto 1);
    inv_in_n_i             : in  std_logic_vector(g_nr_inv_chan downto 1);
    inv_out_o              : out std_logic_vector(g_nr_inv_chan downto 1);
    
    -- Lines for the i2c_slave
    scl_i                  : in  std_logic;
    scl_o                  : out std_logic;
    scl_oe_o               : out std_logic;
    sda_i                  : in  std_logic;
    sda_o                  : out std_logic;
    sda_oe_o               : out std_logic;
    fpga_ga_i              : in  std_logic_vector(4 downto 0);
    fpga_gap_i             : in  std_logic;
--    
--    -- Pins of the SPI interface to write into the Flash memory
--    fpga_prom_cclk_o       : out std_logic;
--    fpga_prom_cso_b_n_o    : out std_logic;
--    fpga_prom_miso_i       : in  std_logic;
--    fpga_prom_mosi_o       : out std_logic;
--      
    fpga_oe_o              : out std_logic;
    fpga_blo_oe_o          : out std_logic;
    fpga_trig_ttl_oe_o     : out std_logic;
    fpga_inv_oe_o          : out std_logic;
    
    -- PLL DAC control pins
    fpga_plldac1_sclk_o    : out std_logic;
    fpga_plldac1_sync_n_o  : out std_logic;
    fpga_plldac1_din_o     : out std_logic;

    -- VCXO DAC control pins
    fpga_plldac2_sclk_o    : out std_logic;
    fpga_plldac2_sync_n_o  : out std_logic;
    fpga_plldac2_din_o     : out std_logic;

    -- SFP I2C lines
    fpga_sfp_mod_def1_b    : inout std_logic;
    fpga_sfp_mod_def2_b    : inout std_logic;

    -- SFP MGT lines
    fpga_mgt_clk0_p_i      : in  std_logic;
    fpga_mgt_clk0_n_i      : in  std_logic;
    mgt_sfp_rx0_p_i        : in  std_logic;
    mgt_sfp_rx0_n_i        : in  std_logic;
    mgt_sfp_tx0_p_o        : out std_logic;
    mgt_sfp_tx0_n_o        : out std_logic;

    -- TTL/INV_TTL_N
    ttl_switch_n_i         : in  std_logic;
    
    extra_switch_n_i       : in  std_logic_vector(7 downto 1);
    
    -- It allows power sequencing of the
    -- 24V rail after a security given delay
    mr_n_o                 : out std_logic;

    -- RTM identifiers, should match with the expected values
    -- TODO: add matching
    fpga_rtmm_n_i          : in std_logic_vector(2 downto 0);
    fpga_rtmp_n_i          : in std_logic_vector(2 downto 0);

    -- DS18B20U+ thermometer bidirectional port
    thermometer_b          : inout std_logic
  );
end conv_ttl_blo_v2;


architecture behav of conv_ttl_blo_v2 is

  --============================================================================
  -- Type declarations
  --============================================================================
  type t_ttl_pulse_counter is array (10 downto 1) 
                              of unsigned(31 downto 0);

  type t_blo_pulse_counter is array (6 downto 1) 
                              of unsigned(31 downto 0);

  --============================================================================
  -- Constant declarations
  --============================================================================
  -- Number of Wishbone masters and slaves, for wb_crossbar
  constant c_nr_masters      : natural :=  1;
  constant c_nr_slaves       : natural := 11;

  -----------------------------------------
  -- Memory map
  -- * all registers are word-addressable
  -- * all registers are word aligned
  -----------------------------------------
  --    PTS           [000-010]
  --    ONEWIRE       [010-020]
  --    DAC_SPI_125   [020-080]
  --    DAC_SPI_20    [080-100]
  --    CLK_INFO_125  [100-120]
  --    CLK_INFO_20   [120-140]
  --    SFP_I2C       [140-200]
  --    ENDPOINT      [200-400]
  --    MINIC         [400-800]
  --    DPRAM         [800-C00]
  --    PULSE_CNT     [C00-FFF]
  -----------------------------------------
  -- slave order definitions
  constant c_slv_pts_regs      : natural :=  0;
  constant c_slv_onewire_mst   : natural :=  1;
  constant c_slv_dac_spi_125   : natural :=  2;
  constant c_slv_dac_spi_20    : natural :=  3;
  constant c_slv_clk_info_125  : natural :=  4;
  constant c_slv_clk_info_20   : natural :=  5;
  constant c_slv_sfp_i2c       : natural :=  6;
  constant c_slv_minic         : natural :=  7;
  constant c_slv_endpoint      : natural :=  8;
  constant c_slv_dpram         : natural :=  9;
  constant c_slv_pulse_cntrs   : natural := 10;

  -- base address definitions
  constant c_addr_pts_regs     : t_wishbone_address := x"00000000";
  constant c_addr_onewire_mst  : t_wishbone_address := x"00000010";
  constant c_addr_dac_spi_125  : t_wishbone_address := x"00000020";
  constant c_addr_dac_spi_20   : t_wishbone_address := x"00000080";
  constant c_addr_clk_info_125 : t_wishbone_address := x"00000100";
  constant c_addr_clk_info_20  : t_wishbone_address := x"00000120";
  constant c_addr_sfp_i2c      : t_wishbone_address := x"00000140";
  constant c_addr_endpoint     : t_wishbone_address := x"00000200";
  constant c_addr_minic        : t_wishbone_address := x"00000400";
  constant c_addr_dpram        : t_wishbone_address := x"00000800";
  constant c_addr_pulse_cntrs  : t_wishbone_address := x"00000C00";

  -- address mask definitions
  constant c_mask_pts_regs     : t_wishbone_address := x"00000FF0"; 
  constant c_mask_onewire_mst  : t_wishbone_address := x"00000FF0"; 
  constant c_mask_dac_spi_125  : t_wishbone_address := x"00000FE0"; 
  constant c_mask_dac_spi_20   : t_wishbone_address := x"00000FE0"; 
  constant c_mask_clk_info_125 : t_wishbone_address := x"00000F60"; 
  constant c_mask_clk_info_20  : t_wishbone_address := x"00000F60";
  constant c_mask_sfp_i2c      : t_wishbone_address := x"00000F60"; 
  constant c_mask_endpoint     : t_wishbone_address := x"00000E00"; 
  constant c_mask_minic        : t_wishbone_address := x"00000C00";
  constant c_mask_dpram        : t_wishbone_address := x"00000C00";
  constant c_mask_pulse_cntrs  : t_wishbone_address := x"00000C00";
  
  -- addresses constant for Wishbone crossbar
  constant c_addresses  : t_wishbone_address_array(c_nr_slaves-1 downto 0)
                        := (
                             c_slv_pts_regs     => c_addr_pts_regs,
                             c_slv_onewire_mst  => c_addr_onewire_mst,
                             c_slv_dac_spi_125  => c_addr_dac_spi_125,
                             c_slv_clk_info_125 => c_addr_clk_info_125,
                             c_slv_dac_spi_20   => c_addr_dac_spi_20,
                             c_slv_clk_info_20  => c_addr_clk_info_20,
                             c_slv_sfp_i2c      => c_addr_sfp_i2c,
                             c_slv_endpoint     => c_addr_endpoint,
                             c_slv_minic        => c_addr_minic,
                             c_slv_dpram        => c_addr_dpram,
                             c_slv_pulse_cntrs  => c_addr_pulse_cntrs
                           );

  -- masks constant for Wishbone crossbar
  constant c_masks      : t_wishbone_address_array(c_nr_slaves-1 downto 0)
                        := (
                             c_slv_pts_regs     => c_mask_pts_regs,
                             c_slv_onewire_mst  => c_mask_onewire_mst,
                             c_slv_dac_spi_125  => c_mask_dac_spi_125,
                             c_slv_clk_info_125 => c_mask_clk_info_125,
                             c_slv_dac_spi_20   => c_mask_dac_spi_20,
                             c_slv_clk_info_20  => c_mask_clk_info_20,
                             c_slv_sfp_i2c      => c_mask_sfp_i2c,
                             c_slv_endpoint     => c_mask_endpoint,
                             c_slv_minic        => c_mask_minic,
                             c_slv_dpram        => c_mask_dpram,
                             c_slv_pulse_cntrs  => c_mask_pulse_cntrs
                           );

  -- MiniNIC log2 of memory size
  constant c_minic_memsize_log2 : natural := 10;

  --============================================================================
  -- Component declarations
  --============================================================================
  -- Reset generator component
  -- (use: global reset generation, output reset generation)
  component reset_gen is
    generic
    (
      -- Reset time in number of clk_i cycles
      g_reset_time : positive := 5_000_000
    );
    port
    (
      clk_i   : in  std_logic;
      rst_i   : in  std_logic;
      rst_n_o : out std_logic
    );
  end component reset_gen;

  -- RTM detector component
  -- (use: detect the presence of an RTM/P module)
  component rtm_detector is
    port 
    (
      rtmm_i    : in std_logic_vector(2 downto 0);
      rtmp_i    : in std_logic_vector(2 downto 0);
      rtmm_ok_o : out std_logic;
      rtmp_ok_o : out std_logic
    );
  end component rtm_detector;

  -- I2C bridge
  -- (use: convert I2C transfers into WB transfers on memmapped registers)  
  component vme64x_i2c is
    port
    (
      -- Clock, reset
      clk_i      : in  std_logic;
      rst_n_i    : in  std_logic;

      -- I2C lines
      sda_en_o   : out std_logic;
      sda_i      : in  std_logic;
      sda_o      : out std_logic;
      scl_en_o   : out std_logic;
      scl_i      : in  std_logic;
      scl_o      : out std_logic;

      -- I2C address and status
      i2c_addr_i : in  std_logic_vector(6 downto 0);
      i2c_done_o : out std_logic;

      -- Wishbone master signals
      wbm_stb_o  : out std_logic;
      wbm_cyc_o  : out std_logic;
      wbm_sel_o  : out std_logic_vector(3 downto 0);
      wbm_we_o   : out std_logic;
      wbm_dat_i  : in  std_logic_vector(31 downto 0);
      wbm_dat_o  : out std_logic_vector(31 downto 0);
      wbm_adr_o  : out std_logic_vector(31 downto 0);
      wbm_ack_i  : in  std_logic;
      wbm_rty_i  : in  std_logic;
      wbm_err_i  : in  std_logic
    );
  end component vme64x_i2c;
  
  -- Regs to test I2C operation
  component pts_regs is
    port (
      rst_n_i                                  : in     std_logic;
      clk_sys_i                                : in     std_logic;
      wb_adr_i                                 : in     std_logic_vector(0 downto 0);
      wb_dat_i                                 : in     std_logic_vector(31 downto 0);
      wb_dat_o                                 : out    std_logic_vector(31 downto 0);
      wb_cyc_i                                 : in     std_logic;
      wb_sel_i                                 : in     std_logic_vector(3 downto 0);
      wb_stb_i                                 : in     std_logic;
      wb_we_i                                  : in     std_logic;
      wb_ack_o                                 : out    std_logic;
      wb_stall_o                               : out    std_logic;
  -- Port for BIT field: 'TTL pulse enable' in reg: 'control and status register'
      pts_csr_ttl_en_o                         : out    std_logic;
  -- Port for BIT field: 'Blocking pulse enable' in reg: 'control and status register'
      pts_csr_blo_en_o                         : out    std_logic;
  -- Port for BIT field: 'Blocking LED control' in reg: 'control and status register'
      pts_csr_blo_led_o                        : out    std_logic;
  -- Port for BIT field: 'pulse LED enable' in reg: 'control and status register'
      pts_csr_pulse_led_en_o                   : out    std_logic;
  -- Port for BIT field: 'status LED enable' in reg: 'control and status register'
      pts_csr_stat_led_en_o                    : out    std_logic;
  -- Port for BIT field: 'reset' in reg: 'control and status register'
      pts_csr_rst_o                            : out    std_logic;
  -- Port for std_logic_vector field: 'switches' in reg: 'control and status register'
      pts_csr_switch_i                         : in     std_logic_vector(7 downto 0);
  -- Port for std_logic_vector field: 'RTM' in reg: 'control and status register'
      pts_csr_rtm_i                            : in     std_logic_vector(5 downto 0);
  -- Port for std_logic_vector field: 'bits' in reg: 'Board ID'
      pts_id_bits_o                            : out    std_logic_vector(31 downto 0)
    );
  end component pts_regs;
  
  -- Fixed-frequency pulse generator component
  -- (use: generate the first pulse that gets replicated from one channel to
  -- another, in the TTL pulse test)
  component ctb_pulse_gen_gp is
    generic
    (
      g_pwidth : natural := 200;
      g_freq   : natural := 400;
      g_delay  : natural := 0
    );
    port
    (
      clk_i   : in  std_logic;
      rst_n_i : in  std_logic;
      en_i    : in  std_logic;
      pulse_o : out std_logic
    );
  end component ctb_pulse_gen_gp;

  -- Pulse generator component
  -- (use: TTL and BLO pulse tests)
  component ctb_pulse_gen is
    generic 
    (
      -- Pulse width, in number of clk_i cycles
      g_pulse_width     : natural := 15;

      -- Glitch filter length:
      -- g_glitch_filt_len=1 => trigger width should be > 1 clk_i cycle
      -- g_glitch_filt_len=2 => trigger width should be > 2 clk_i cycles
      -- etc.
      g_glitch_filt_len : natural := 4

    );
    port
    (
      clk_i        : in  std_logic;
      rst_n_i      : in  std_logic;

      -- Pulse type input
      -- '1' - TYPE 1 (glitch-sensitive, without output jitter)
      -- '0' - TYPE 2 (glitch-insensitive, with output jitter)
      pulse_type_i : in  std_logic;

      -- Enable input, pulse generation is enabled when '1'
      en_i         : in  std_logic;

      -- Trigger input, has to be '1' to assure pulse output with delay no greater
      -- than internal gate delays.
      trig_i       : in  std_logic;

      -- Pulse output, active-high
      -- latency:
      --    TYPE 1 pulse: none
      --    TYPE 2 pulse: g_glitch_filt_len+3 clk_i cycles
      pulse_o      : out std_logic
    );
  end component ctb_pulse_gen;
  
  -- Pulse counter Wishbone regs component
  -- use: TTL, BLO tests
  component pulse_cnt_wb is
    port
    (
      rst_n_i                                  : in     std_logic;
      clk_sys_i                                : in     std_logic;
      wb_adr_i                                 : in     std_logic_vector(4 downto 0);
      wb_dat_i                                 : in     std_logic_vector(31 downto 0);
      wb_dat_o                                 : out    std_logic_vector(31 downto 0);
      wb_cyc_i                                 : in     std_logic;
      wb_sel_i                                 : in     std_logic_vector(3 downto 0);
      wb_stb_i                                 : in     std_logic;
      wb_we_i                                  : in     std_logic;
      wb_ack_o                                 : out    std_logic;
      wb_stall_o                               : out    std_logic;
      pulse_cnt_ch1o_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch1i_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch2o_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch2i_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch3o_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch3i_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch4o_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch4i_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch5o_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch5i_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch6o_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch6i_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch7o_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch7i_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch8o_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch8i_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch9o_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch9i_val_i                     : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch10o_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch10i_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch11o_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch11i_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch12o_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch12i_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch13o_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch13i_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch14o_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch14i_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch15o_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch15i_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch16o_val_i                    : in     std_logic_vector(31 downto 0);
      pulse_cnt_ch16i_val_i                    : in     std_logic_vector(31 downto 0)
    );
  end component pulse_cnt_wb;

  -- Incremental counter component
  -- use: DAC, PLL & VCXO test
  component incr_counter is
    generic
    (
      width             : integer := 32 -- default size
    );
    port
    (
      -- INPUTS
      -- Signals from the clk_rst_manager
      clk_i             : in std_logic;
      rst_i             : in std_logic;

      -- Signals from any unit
      counter_top_i     : in std_logic_vector(width-1 downto 0); -- max value to be counted; when reached
                                                                 -- counter stays at it, until a reset
      counter_incr_en_i : in std_logic;   -- enables counting

     -- OUTPUTS
     -- Signals to any unit
      counter_o         : out std_logic_vector(width-1 downto 0);
      counter_is_full_o : out std_logic
    ); -- counter reahed counter_top_i value
  end component incr_counter;

  -- Clock info component
  -- use: DAC, PLL & VCXO test
  component clk_info_wb_slave is
    port
    (
      -- WISHBONE slave signals
      wb_clk_i          : in  std_logic;
      rst_i             : in  std_logic;
      wb_cyc_i          : in  std_logic;
      wb_stb_i          : in  std_logic;
      wb_addr_i         : in  std_logic_vector( 2 downto 0);
      wb_data_i         : in  std_logic_vector(31 downto 0);
      wb_we_i           : in  std_logic;
      wb_data_o         : out std_logic_vector(31 downto 0);
      wb_ack_o          : out std_logic;
       -- Clock info signals
      counter_is_full_i : in std_logic;
      counter_now_i     : in std_logic_vector (31 downto 0);
      counter_top_o     : out std_logic_vector (31 downto 0);
      counter_rst_o     : out std_logic;
      counter_en_o      : out std_logic;
      oe_clk_o          : out std_logic
    );
  end component clk_info_wb_slave;

  -- Mini NIC component
  -- use: MGTX and SFP test
  component xwr_mini_nic
    generic (
      g_interface_mode       : t_wishbone_interface_mode;
      g_address_granularity  : t_wishbone_address_granularity;
      g_memsize_log2         : integer;
      g_buffer_little_endian : boolean);
    port (
      clk_sys_i           : in  std_logic;
      rst_n_i             : in  std_logic;
      mem_data_o          : out std_logic_vector(31 downto 0);
      mem_addr_o          : out std_logic_vector(g_memsize_log2-1 downto 0);
      mem_data_i          : in  std_logic_vector(31 downto 0);
      mem_wr_o            : out std_logic;
      src_o               : out t_wrf_source_out;
      src_i               : in  t_wrf_source_in;
      snk_o               : out t_wrf_sink_out;
      snk_i               : in  t_wrf_sink_in;
      txtsu_port_id_i     : in  std_logic_vector(4 downto 0);
      txtsu_frame_id_i    : in  std_logic_vector(16 - 1 downto 0);
      txtsu_tsval_i       : in  std_logic_vector(28 + 4 - 1 downto 0);
      txtsu_tsincorrect_i : in  std_logic;
      txtsu_stb_i         : in  std_logic;
      txtsu_ack_o         : out std_logic;
      wb_i                : in  t_wishbone_slave_in;
      wb_o                : out t_wishbone_slave_out);
  end component;
  
  -- GTP component for Spartan-6 FPGA
  -- use: MGTX and SFP test
  component wr_gtp_phy_spartan6
    generic (
      g_simulation      : integer;
      g_force_disparity : integer;
      g_enable_ch0      : integer;
      g_enable_ch1      : integer);
    port (
      gtp_clk_i          : in  std_logic;
      ch0_ref_clk_i      : in  std_logic                    := '0';
      ch0_tx_data_i      : in  std_logic_vector(7 downto 0) := "00000000";
      ch0_tx_k_i         : in  std_logic                    := '0';
      ch0_tx_disparity_o : out std_logic;
      ch0_tx_enc_err_o   : out std_logic;
      ch0_rx_rbclk_o     : out std_logic;
      ch0_rx_data_o      : out std_logic_vector(7 downto 0);
      ch0_rx_k_o         : out std_logic;
      ch0_rx_enc_err_o   : out std_logic;
      ch0_rx_bitslide_o  : out std_logic_vector(3 downto 0);
      ch0_rst_i          : in  std_logic                    := '0';
      ch0_loopen_i       : in  std_logic                    := '0';
      ch1_ref_clk_i      : in  std_logic                    := '0';
      ch1_tx_data_i      : in  std_logic_vector(7 downto 0) := "00000000";
      ch1_tx_k_i         : in  std_logic                    := '0';
      ch1_tx_disparity_o : out std_logic;
      ch1_tx_enc_err_o   : out std_logic;
      ch1_rx_data_o      : out std_logic_vector(7 downto 0);
      ch1_rx_rbclk_o     : out std_logic;
      ch1_rx_k_o         : out std_logic;
      ch1_rx_enc_err_o   : out std_logic;
      ch1_rx_bitslide_o  : out std_logic_vector(3 downto 0);
      ch1_rst_i          : in  std_logic                    := '0';
      ch1_loopen_i       : in  std_logic                    := '0';
      pad_txn0_o         : out std_logic;
      pad_txp0_o         : out std_logic;
      pad_rxn0_i         : in  std_logic                    := '0';
      pad_rxp0_i         : in  std_logic                    := '0';
      pad_txn1_o         : out std_logic;
      pad_txp1_o         : out std_logic;
      pad_rxn1_i         : in  std_logic                    := '0';
      pad_rxp1_i         : in  std_logic                    := '0');
  end component;

  --============================================================================
  -- Signal declarations
  --============================================================================
  -- Clock signals
  signal clk125              : std_logic;

  -- Global reset signals
  signal rst_n, rst          : std_logic;

  -- Temporary signal for inverted-TTL pulse outputs
  signal inv_outputs         : std_logic_vector(g_nr_inv_chan downto 1);
  
  -- Output enable signals
  signal oe, ttl_oe          : std_logic;
  signal blo_oe, inv_oe      : std_logic;

  -- Signal for controlling the bicolor LED matrix
  signal bicolor_led_state   : std_logic_vector(23 downto 0);

  -- Wishbone crossbar signals
  signal xbar_slave_in       : t_wishbone_slave_in_array   (c_nr_masters - 1 downto 0);
  signal xbar_slave_out      : t_wishbone_slave_out_array  (c_nr_masters - 1 downto 0);
  signal xbar_master_in      : t_wishbone_master_in_array  (c_nr_slaves  - 1 downto 0);
  signal xbar_master_out     : t_wishbone_master_out_array (c_nr_slaves  - 1 downto 0);
  
  -- I2C bridge signals
  signal i2c_done            : std_logic;
  signal i2c_up              : std_logic;
  signal i2c_addr            : std_logic_vector(6 downto 0);

  -- LED signals
  signal led_seq             : unsigned(4 downto 0);
  signal pulse_led_en        : std_logic;
  signal stat_led_en         : std_logic;
  
  -- PTS register signals
  signal pts_state           : t_pts_state;
  signal cnt_halfsec         : unsigned(25 downto 0);
  signal rst_fr_reg          : std_logic;

  -- TTL pulse test signals
  signal ttl_pulse_en        : std_logic;

  signal ttl_trigs           : std_logic_vector(10 downto 1);
  signal ttl_trigs_d0        : std_logic_vector(10 downto 1);
  signal ttl_trigs_d1        : std_logic_vector(10 downto 1);
  signal ttl_trigs_d2        : std_logic_vector(10 downto 1);
  signal ttl_pulses          : std_logic_vector(10 downto 1);
  signal ttl_pulses_d0       : std_logic_vector(10 downto 1);
  signal ttl_pulses_d1       : std_logic_vector(10 downto 1);
  signal ttl_pulses_d2       : std_logic_vector(10 downto 1);

  signal cnt_ttl_in          : t_ttl_pulse_counter;
  signal cnt_ttl_out         : t_ttl_pulse_counter;

  -- Blocking pulse test signals
  signal blo_pulse_en        : std_logic;

  signal blo_led             : std_logic;

  signal blo_trigs           : std_logic_vector(6 downto 1);
  signal blo_trigs_d0        : std_logic_vector(6 downto 1);
  signal blo_trigs_d1        : std_logic_vector(6 downto 1);
  signal blo_trigs_d2        : std_logic_vector(6 downto 1);
  signal blo_pulses          : std_logic_vector(6 downto 1);
  signal blo_pulses_d0       : std_logic_vector(6 downto 1);
  signal blo_pulses_d1       : std_logic_vector(6 downto 1);
  signal blo_pulses_d2       : std_logic_vector(6 downto 1);

  signal cnt_blo_in          : t_blo_pulse_counter;
  signal cnt_blo_out         : t_blo_pulse_counter;

  signal rtm_lines           : std_logic_vector(5 downto 0);
  signal switches            : std_logic_vector(7 downto 0);

  signal dumled              : std_logic_vector(6 downto 1);

  -- one-wire master signals
  signal owr_pwren           : std_logic_vector(0 downto 0);
  signal owr_en              : std_logic_vector(0 downto 0);
  signal owr_in              : std_logic_vector(0 downto 0);

  -- PLL & DAC test signals
  signal cnt125              : std_logic_vector(31 downto 0);
  signal cnt125_top          : std_logic_vector(31 downto 0);
  signal cnt125_en           : std_logic;
  signal cnt125_rst          : std_logic;
  signal cnt125_full         : std_logic;
  signal cnt125_actual_rst   : std_logic;
  signal plldac1_sync_n      : std_logic_vector(7 downto 0);
  
  signal cnt20               : std_logic_vector(31 downto 0);
  signal cnt20_top           : std_logic_vector(31 downto 0);
  signal cnt20_en            : std_logic;
  signal cnt20_rst           : std_logic;
  signal cnt20_full          : std_logic;
  signal cnt20_actual_rst    : std_logic;
  signal cnt20_actual_rst_d0 : std_logic;
  signal plldac2_sync_n      : std_logic_vector(7 downto 0);

  -- SFP EEPROM signals
  signal i2c_scl_fr_sfp      : std_logic;
  signal i2c_scl_to_sfp      : std_logic;
  signal i2c_sfp_scl_en      : std_logic;

  signal i2c_sda_fr_sfp      : std_logic;
  signal i2c_sda_to_sfp      : std_logic;
  signal i2c_sfp_sda_en      : std_logic;

  -- SFP signals
  signal clk_gtp             : std_logic;
  
  signal minic_mem_data_out  : std_logic_vector(31 downto 0);
  signal minic_mem_addr_out  : std_logic_vector(c_minic_memsize_log2-1 downto 0);
  signal minic_mem_data_in   : std_logic_vector(31 downto 0);
  signal minic_mem_wr        : std_logic;
  signal minic_src_out       : t_wrf_source_out;
  signal minic_src_in        : t_wrf_source_in;
  signal minic_snk_out       : t_wrf_sink_out;
  signal minic_snk_in        : t_wrf_sink_in;
  
  signal phy_rst             : std_logic;
  signal phy_loopen          : std_logic;
  signal phy_tx_d            : std_logic_vector(15 downto 0);
  signal phy_tx_k            : std_logic_vector( 1 downto 0);
  signal phy_tx_disparity    : std_logic;
  signal phy_tx_error        : std_logic;
  signal phy_rx_d            : std_logic_vector(15 downto 0);
  signal phy_rx_clk          : std_logic;
  signal phy_rx_k            : std_logic_vector( 1 downto 0);
  signal phy_rx_error        : std_logic;
  signal phy_rx_bitslide     : std_logic_vector( 4 downto 0);

  signal dpram_we            : std_logic;

begin

  --============================================================================
  -- Generate 125 MHz global signal from differential lines
  --============================================================================
  cmp_125_diff_buf : IBUFDS
    generic map 
    (
      DIFF_TERM    => TRUE,
      IBUF_LOW_PWR => TRUE
    )
    port map 
    (
      I  => fpga_clk_p_i,
      IB => fpga_clk_n_i,
      O  => clk125
    );

  --============================================================================
  -- Internal and external reset generation
  --============================================================================
  -- Configure reset generator for 96ms power-on reset
  cmp_reset_gen : reset_gen
    generic map
    (
      -- Reset time: 12 * 8ns * (10**6) = 96 ms
      g_reset_time => 12*(10**6)
    )
    port map
    (
      clk_i   => clk125,
      rst_i   => rst_fr_reg,
      rst_n_o => rst_n
    );
  
  rst    <= not rst_n;
  mr_n_o <= rst_n;
  
  --============================================================================
  -- I2C bridge logic
  --============================================================================
  i2c_addr <= "10" & fpga_ga_i;
  
  cmp_i2c_bridge : vme64x_i2c
    port map
    (
      -- Clock, reset
      clk_i      => clk125,
      rst_n_i    => rst_n,

      -- I2C lines
      sda_en_o   => sda_oe_o,
      sda_i      => sda_i,
      sda_o      => sda_o,
      scl_en_o   => scl_oe_o,
      scl_i      => scl_i,
      scl_o      => scl_o,

      -- I2C address and status
      i2c_addr_i => i2c_addr,
      i2c_done_o => i2c_done,

      -- Wishbone master signals
      wbm_stb_o  => xbar_slave_in(0).stb,
      wbm_cyc_o  => xbar_slave_in(0).cyc,
      wbm_sel_o  => xbar_slave_in(0).sel,
      wbm_we_o   => xbar_slave_in(0).we,
      wbm_dat_i  => xbar_slave_out(0).dat,
      wbm_dat_o  => xbar_slave_in(0).dat,
      wbm_adr_o  => xbar_slave_in(0).adr,
      wbm_ack_i  => xbar_slave_out(0).ack,
      wbm_rty_i  => xbar_slave_out(0).rty,
      wbm_err_i  => xbar_slave_out(0).err
    );
  
  -- Process to set the I2C_UP signal for display on the front panel
  -- of the front module. The I2C_UP signal is permanently set once an
  -- I2C transfer has successfully completed, as signaled by the RD_DONE
  -- and WR_DONE outputs of the I2C slave.
  p_i2c_up : process (clk125) is
  begin
    if rising_edge(clk125) then
      if (rst_n = '0') then
        i2c_up <= '0';
      elsif (i2c_done = '1') then
        i2c_up <= '1';
      end if;
    end if;
  end process p_i2c_up;

  --============================================================================
  -- Instantiation and connection of the main Wishbone crossbar
  --============================================================================
  xbar_master_in(0).int <= '0';
  xbar_master_in(0).err <= '0';

  cmp_wb_crossbar : xwb_crossbar
  generic map
  (
    g_num_masters => c_nr_masters,
    g_num_slaves  => c_nr_slaves,
    g_registered  => false,
    g_address     => c_addresses,
    g_mask        => c_masks
  )
  port map
  (
    clk_sys_i => clk125,
    rst_n_i   => rst_n,
    slave_i   => xbar_slave_in,
    slave_o   => xbar_slave_out,
    master_i  => xbar_master_in,
    master_o  => xbar_master_out
  );

  --============================================================================
  -- PTS registers
  --============================================================================
  -- Assign the RTM detection lines and switches signals
  rtm_lines <= fpga_rtmp_n_i & fpga_rtmm_n_i;
  switches  <= ttl_switch_n_i & extra_switch_n_i;
  
  -- Regs to test I2C operation
  cmp_pts_regs : pts_regs
    port map
    (
      rst_n_i        => rst_n,
      clk_sys_i      => clk125,
      wb_adr_i       => xbar_master_out(c_slv_pts_regs).adr(2 downto 2),
      wb_dat_i       => xbar_master_out(c_slv_pts_regs).dat,
      wb_dat_o       => xbar_master_in(c_slv_pts_regs).dat,
      wb_cyc_i       => xbar_master_out(c_slv_pts_regs).cyc,
      wb_sel_i       => xbar_master_out(c_slv_pts_regs).sel,
      wb_stb_i       => xbar_master_out(c_slv_pts_regs).stb,
      wb_we_i        => xbar_master_out(c_slv_pts_regs).we,
      wb_ack_o       => xbar_master_in(c_slv_pts_regs).ack,
      wb_stall_o     => xbar_master_in(c_slv_pts_regs).stall,

      -- PTS control register
      pts_csr_ttl_en_o       => ttl_pulse_en,
      pts_csr_blo_en_o       => blo_pulse_en,
      pts_csr_blo_led_o      => blo_led,
      pts_csr_pulse_led_en_o => pulse_led_en,
      pts_csr_stat_led_en_o  => stat_led_en,
      pts_csr_rst_o          => rst_fr_reg,
      pts_csr_switch_i       => switches,
      pts_csr_rtm_i          => rtm_lines,

      -- PTS ID register
      pts_id_bits_o => open
    );

  --============================================================================
  -- Thermometer test logic
  --============================================================================
  -- The one-wire master component is used to control the on-board DS18B20
  -- thermometer
  cmp_onewire_master : wb_onewire_master
    generic map
    (
      g_interface_mode      => CLASSIC,
      g_address_granularity => WORD,
      g_num_ports           => 1,
      g_ow_btp_normal       => "5.0",
      g_ow_btp_overdrive    => "1.0"
    )
    port map
    (
      clk_sys_i   => clk125,
      rst_n_i     => rst_n,

      wb_cyc_i    => xbar_master_out(c_slv_onewire_mst).cyc,
      wb_sel_i    => xbar_master_out(c_slv_onewire_mst).sel,
      wb_stb_i    => xbar_master_out(c_slv_onewire_mst).stb,
      wb_we_i     => xbar_master_out(c_slv_onewire_mst).we,
      wb_adr_i    => xbar_master_out(c_slv_onewire_mst).adr(4 downto 2), 
      wb_dat_i    => xbar_master_out(c_slv_onewire_mst).dat,
      wb_dat_o    => xbar_master_in(c_slv_onewire_mst).dat,
      wb_ack_o    => xbar_master_in(c_slv_onewire_mst).ack, 
      wb_int_o    => open, 
      wb_stall_o  => xbar_master_in(c_slv_onewire_mst).stall,

      owr_pwren_o => owr_pwren,
      owr_en_o    => owr_en,
      owr_i       => owr_in
    );
     
  -- Generate tri-state buffer for thermometer
  thermometer_b <= '0' when (owr_en(0) = '1') else 
                   'Z';
  owr_in(0)     <= thermometer_b;

  --============================================================================
  -- Pulse output enable logic
  --============================================================================
  -- The general output enable is set first and the blocking, TTL
  -- and INV output enable signals are set one clock cycle later.
  p_oe : process(clk125)
  begin
    if rising_edge(clk125) then
      if (rst_n = '0') then
        oe     <= '0';
        blo_oe <= '0';
        ttl_oe <= '0';
        inv_oe <= '0';
      else
        oe <= '1';
        if (oe = '1') then
          blo_oe <= '1';
          ttl_oe <= '1';
          inv_oe <= '1';
        end if;
      end if;
    end if;
  end process p_oe;

  -- Assign output enable registers to chip outputs
  fpga_oe_o          <= oe;
  fpga_blo_oe_o      <= blo_oe;
  fpga_trig_ttl_oe_o <= ttl_oe;
  fpga_inv_oe_o      <= inv_oe;

  --============================================================================
  -- TTL pulse test logic
  --============================================================================
  -- First, instantiate a general-purpose pulse generator to generate the output 
  -- pulse from CH10 (INV-TTL CH4) to CH1
  --
  -- 1-us pulses are generated twice a second.
  cmp_first_pulse_gen : ctb_pulse_gen_gp
    generic map
    (
      g_pwidth => 125,
      g_freq   => 62*(10**6),
      g_delay  => 0
    )
    port map
    (
      clk_i    => clk125,
      rst_n_i  => rst_n,
      en_i     => ttl_pulse_en,
      pulse_o  => ttl_pulses(10)
    );

  -- Assign the trigger inputs to internal signals
  ttl_trigs( 6 downto 1) <= not fpga_input_ttl_n_i;
  ttl_trigs(10 downto 7) <= not inv_in_n_i;

  -- Now, generate nine pulse generator blocks connected to the TTL outputs
  -- and with the TTL inputs as triggers. External to the FPGA, the inputs of
  -- CH2 are expected to be connected to CH1, CH3 to CH2 and so on, until
  -- the last INV_TTL output, which is expected to be connected back to the
  -- input of CH1. This last INV-TTL output is connected above to the
  -- fixed-frequency pulse generator.
  --
  -- Type 1 pulses (non-glich-filtered) with 1us width will be generated by 
  -- the generator blocks.
  gen_ttl_pulse_gens : for i in 1 to 9 generate
    cmp_ttl_pulse_gen : ctb_pulse_gen
      generic map
      (
        g_pulse_width     => 125,
        g_glitch_filt_len => 4
      )
      port map
      (
        clk_i        => clk125,
        rst_n_i      => rst_n,
        pulse_type_i => '1',
        en_i         => ttl_pulse_en,
        trig_i       => ttl_trigs(i),
        pulse_o      => ttl_pulses(i)
      );
  end generate gen_ttl_pulse_gens;

  -- Assign the outputs of the ctb_pulse_gen components to the TTL and INV-TTL
  -- outputs
  fpga_out_ttl_o <= ttl_pulses( 6 downto 1);
  inv_out_o      <= ttl_pulses(10 downto 7);

  -- Process to count input and output pulses. Since the pulses are generated
  -- on the rising edge of the input pulse, the outputs from the ctb_pulse_gen
  -- blocks need to  be resynced.
  p_cnt_ttl_pulses : process(clk125) is
  begin
    if rising_edge(clk125) then
      if (rst_n = '0') then
        cnt_ttl_in    <= (others => (others => '0'));
        cnt_ttl_out   <= (others => (others => '0'));
        ttl_trigs_d0  <= (others => '0');
        ttl_trigs_d1  <= (others => '0');
        ttl_trigs_d2  <= (others => '0');
        ttl_pulses_d0 <= (others => '0');
        ttl_pulses_d1 <= (others => '0');
        ttl_pulses_d2 <= (others => '0');
      else
        -- Resync each channel input and output pulse and increment the
        -- corresponding counter on rising edge of the pulse signal
        for i in 1 to 10 loop
          ttl_trigs_d0(i) <= ttl_trigs(i);
          ttl_trigs_d1(i) <= ttl_trigs_d0(i);
          ttl_trigs_d2(i) <= ttl_trigs_d1(i);
          if (ttl_trigs_d1(i) = '1') and (ttl_trigs_d2(i) = '0') then
            cnt_ttl_in(i) <= cnt_ttl_in(i) + 1;
          end if;

          ttl_pulses_d0(i) <= ttl_pulses(i);
          ttl_pulses_d1(i) <= ttl_pulses_d0(i);
          ttl_pulses_d2(i) <= ttl_pulses_d1(i);
          if (ttl_pulses_d1(i) = '1') and (ttl_pulses_d2(i) = '0') then
            cnt_ttl_out(i) <= cnt_ttl_out(i) + 1;
          end if;
        end loop;
      end if;
    end if;
  end process p_cnt_ttl_pulses;

  --============================================================================
  -- Blocking pulse test logic
  --============================================================================
  -- First, instantiate six fixed-frequency, fixed-delay pulse generators
  -- (ctb_pulse_gen_gp components) to output the pulses on each channel. The channels
  -- are time-domain multiplexed, each channel outputting a pulse 100 ms after
  -- the other. Blocking CH1 will output a pulse at 0 ms, CH2 at 100 ms, CH3 at
  -- 200 ms, and so on.
  gen_blo_pulse_gen : for i in 1 to 6 generate
    cmp_blo_pulse_gen : ctb_pulse_gen_gp
      generic map
      (
        g_pwidth => 125,
        g_freq   => 75*(10**6),
        g_delay  => (i-1)*125*(10**5)
      )
      port map
      (
        clk_i   => clk125,
        rst_n_i => rst_n,
        en_i    => blo_pulse_en,
        pulse_o => blo_pulses(i)
      );
  end generate gen_blo_pulse_gen;

  -- Then, assign the blocking inputs to the blocking "trigger" signals and the
  -- blocking pulses to blocking outputs
  blo_trigs       <= fpga_blo_in_i;
  fpga_trig_blo_o <= blo_pulses;

  -- Process to count input and output pulses. Since the pulses are generated
  -- on the rising edge of the input pulse, the outputs from the ctb_pulse_gen
  -- blocks need to  be resynced.
  p_cnt_blo_pulses : process(clk125) is
  begin
    if rising_edge(clk125) then
      if (rst_n = '0') then
        cnt_blo_in    <= (others => (others => '0'));
        cnt_blo_out   <= (others => (others => '0'));
        blo_trigs_d0  <= (others => '0');
        blo_trigs_d1  <= (others => '0');
        blo_trigs_d2  <= (others => '0');
        blo_pulses_d0 <= (others => '0');
        blo_pulses_d1 <= (others => '0');
        blo_pulses_d2 <= (others => '0');
      else
        -- Resync each channel input and output pulse and increment the
        -- corresponding counter on rising edge of the pulse signal
        for i in 1 to 6 loop
          blo_trigs_d0(i) <= blo_trigs(i);
          blo_trigs_d1(i) <= blo_trigs_d0(i);
          blo_trigs_d2(i) <= blo_trigs_d1(i);
          if (blo_trigs_d1(i) = '1') and (blo_trigs_d2(i) = '0') then
            cnt_blo_in(i) <= cnt_blo_in(i) + 1;
          end if;

          blo_pulses_d0(i) <= blo_pulses(i);
          blo_pulses_d1(i) <= blo_pulses_d0(i);
          blo_pulses_d2(i) <= blo_pulses_d1(i);
          if (blo_pulses_d1(i) = '1') and (blo_pulses_d2(i) = '0') then
            cnt_blo_out(i) <= cnt_blo_out(i) + 1;
          end if;
        end loop;
      end if;
    end if;
  end process p_cnt_blo_pulses;

  -- Pulse LEDs are connected back to RTM detection lines, so here we connect
  -- the output to the blocking pulse LED control bit
  pulse_rear_led_n_o <= (others => blo_led);

  -- Uncomment following lines to debug BLO pulse output
  -- NOTE: you need a blocking RTM for this to work properly.
  -- gen_pulse_leds : for i in 1 to 6 generate
  --   cmp_pulse_led_gen : ctb_pulse_gen_gp
  --     generic map
  --     (
  --       g_pwidth => 12*(10**6),
  --       g_freq   => 75*(10**6),
  --       g_delay  => (i-1)*125*(10**5)
  --     )
  --     port map
  --     (
  --       clk_i   => clk125,
  --       rst_n_i => rst_n,
  --       en_i    => '1',
  --       pulse_o => dumled(i)
  --     );
  -- end generate gen_pulse_leds;

  -- pulse_rear_led_n_o  <= not dumled;

  --============================================================================
  -- Pulse counter registers, holding values for pulse counters of both blocking
  -- and TTL pulse repetition tests.
  --============================================================================
  cmp_pulse_cnt_wb : pulse_cnt_wb
    port map
    (
      rst_n_i               => rst_n,
      clk_sys_i             => clk125,
      wb_adr_i              => xbar_master_out(c_slv_pulse_cntrs).adr(6 downto 2),
      wb_dat_i              => xbar_master_out(c_slv_pulse_cntrs).dat,
      wb_dat_o              => xbar_master_in(c_slv_pulse_cntrs).dat,
      wb_cyc_i              => xbar_master_out(c_slv_pulse_cntrs).cyc,
      wb_sel_i              => xbar_master_out(c_slv_pulse_cntrs).sel,
      wb_stb_i              => xbar_master_out(c_slv_pulse_cntrs).stb,
      wb_we_i               => xbar_master_out(c_slv_pulse_cntrs).we,
      wb_ack_o              => xbar_master_in(c_slv_pulse_cntrs).ack,
      wb_stall_o            => xbar_master_in(c_slv_pulse_cntrs).stall,
    
      pulse_cnt_ch1o_val_i  => std_logic_vector(cnt_ttl_out( 1)),
      pulse_cnt_ch1i_val_i  => std_logic_vector(cnt_ttl_in ( 1)),
      pulse_cnt_ch2o_val_i  => std_logic_vector(cnt_ttl_out( 2)),
      pulse_cnt_ch2i_val_i  => std_logic_vector(cnt_ttl_in ( 2)),
      pulse_cnt_ch3o_val_i  => std_logic_vector(cnt_ttl_out( 3)),
      pulse_cnt_ch3i_val_i  => std_logic_vector(cnt_ttl_in ( 3)),
      pulse_cnt_ch4o_val_i  => std_logic_vector(cnt_ttl_out( 4)),
      pulse_cnt_ch4i_val_i  => std_logic_vector(cnt_ttl_in ( 4)),
      pulse_cnt_ch5o_val_i  => std_logic_vector(cnt_ttl_out( 5)),
      pulse_cnt_ch5i_val_i  => std_logic_vector(cnt_ttl_in ( 5)),
      pulse_cnt_ch6o_val_i  => std_logic_vector(cnt_ttl_out( 6)),
      pulse_cnt_ch6i_val_i  => std_logic_vector(cnt_ttl_in ( 6)),
      pulse_cnt_ch7o_val_i  => std_logic_vector(cnt_ttl_out( 7)),
      pulse_cnt_ch7i_val_i  => std_logic_vector(cnt_ttl_in ( 7)),
      pulse_cnt_ch8o_val_i  => std_logic_vector(cnt_ttl_out( 8)),
      pulse_cnt_ch8i_val_i  => std_logic_vector(cnt_ttl_in ( 8)),
      pulse_cnt_ch9o_val_i  => std_logic_vector(cnt_ttl_out( 9)),
      pulse_cnt_ch9i_val_i  => std_logic_vector(cnt_ttl_in ( 9)),
      pulse_cnt_ch10o_val_i => std_logic_vector(cnt_ttl_out(10)),
      pulse_cnt_ch10i_val_i => std_logic_vector(cnt_ttl_in (10)),
      pulse_cnt_ch11o_val_i => std_logic_vector(cnt_blo_out( 1)),
      pulse_cnt_ch11i_val_i => std_logic_vector(cnt_blo_in ( 1)),
      pulse_cnt_ch12o_val_i => std_logic_vector(cnt_blo_out( 2)),
      pulse_cnt_ch12i_val_i => std_logic_vector(cnt_blo_in ( 2)),
      pulse_cnt_ch13o_val_i => std_logic_vector(cnt_blo_out( 3)),
      pulse_cnt_ch13i_val_i => std_logic_vector(cnt_blo_in ( 3)),
      pulse_cnt_ch14o_val_i => std_logic_vector(cnt_blo_out( 4)),
      pulse_cnt_ch14i_val_i => std_logic_vector(cnt_blo_in ( 4)),
      pulse_cnt_ch15o_val_i => std_logic_vector(cnt_blo_out( 5)),
      pulse_cnt_ch15i_val_i => std_logic_vector(cnt_blo_in ( 5)),
      pulse_cnt_ch16o_val_i => std_logic_vector(cnt_blo_out( 6)),
      pulse_cnt_ch16i_val_i => std_logic_vector(cnt_blo_in ( 6))
    );

  --============================================================================
  -- PLL test logic
  -- * test OSC3 -- 20MHz VCXO
  -- * test IC13 -- 125MHz PLL from TI
  -- * test IC17, IC18 -- Analog Devices AD5662 DACs used to control operation 
  --   of OSC1 & OSC3
  --============================================================================
  -----------------------------------------------------------------------------
  -- 125-MHz PLL test
  -----------------------------------------------------------------------------
  -- 125-MHz counter reset signal is OR between global reset and register reset
  cnt125_actual_rst <= cnt125_rst or rst;

  -- Instantiate increment counter for IC13
  cmp_pll125_incr_counter : incr_counter
    generic map
    (
      width => 32
    )
    port map
    (
      clk_i             => clk125,
      rst_i             => cnt125_actual_rst,
      counter_top_i     => cnt125_top,
      counter_incr_en_i => cnt125_en,
      counter_o         => cnt125,
      counter_is_full_o => cnt125_full
    );

  -- Instantiate clock info slave
  cmp_pll125_clk_info : clk_info_wb_slave
    port map
    (
      wb_clk_i          => clk125,
      rst_i             => rst,

      wb_cyc_i          => xbar_master_out(c_slv_clk_info_125).cyc,
      wb_stb_i          => xbar_master_out(c_slv_clk_info_125).stb,
      wb_addr_i         => xbar_master_out(c_slv_clk_info_125).adr(4 downto 2),
      wb_data_i         => xbar_master_out(c_slv_clk_info_125).dat,
      wb_we_i           => xbar_master_out(c_slv_clk_info_125).we,
      wb_data_o         => xbar_master_in(c_slv_clk_info_125).dat,
      wb_ack_o          => xbar_master_in(c_slv_clk_info_125).ack, 

      counter_is_full_i => cnt125_full,
      counter_now_i     => cnt125,
      counter_top_o     => cnt125_top,
      counter_rst_o     => cnt125_rst,
      counter_en_o      => cnt125_en,
      oe_clk_o          => open
    );

  -- Instantate a Wishbone SPI module to control IC18 DAC operation
  cmp_dac_125_spi : wb_spi
    generic map
    (
      g_interface_mode      => CLASSIC,
      g_address_granularity => WORD
    )
    port map
    (
      clk_sys_i  => clk125,
      rst_n_i    => rst_n,

      wb_adr_i   => xbar_master_out(c_slv_dac_spi_125).adr(6 downto 2),
      wb_dat_i   => xbar_master_out(c_slv_dac_spi_125).dat,
      wb_dat_o   => xbar_master_in(c_slv_dac_spi_125).dat,
      wb_sel_i   => xbar_master_out(c_slv_dac_spi_125).sel,
      wb_stb_i   => xbar_master_out(c_slv_dac_spi_125).stb,
      wb_cyc_i   => xbar_master_out(c_slv_dac_spi_125).cyc,
      wb_we_i    => xbar_master_out(c_slv_dac_spi_125).we,
      wb_ack_o   => xbar_master_in(c_slv_dac_spi_125).ack,
      wb_err_o   => xbar_master_in(c_slv_dac_spi_125).err,
      wb_int_o   => xbar_master_in(c_slv_dac_spi_125).int,
      wb_stall_o => xbar_master_in(c_slv_dac_spi_125).stall,

      pad_cs_o   => plldac1_sync_n,
      pad_sclk_o => fpga_plldac1_sclk_o,
      pad_mosi_o => fpga_plldac1_din_o,
      pad_miso_i => '0'
    );
  
  -- Finally, assign the SYNC_N output to the DAC
  fpga_plldac1_sync_n_o <= plldac1_sync_n(0);

  -----------------------------------------------------------------------------
  -- 20-MHz VCXO test
  -----------------------------------------------------------------------------
  -- First, synchronize the reset signal into the 20 MHz clock domain
  p_sync_rst : process (clk20_vcxo_i)
  begin
    if rising_edge(clk20_vcxo_i) then
      cnt20_actual_rst_d0 <= cnt20_rst;
      cnt20_actual_rst    <= cnt20_actual_rst_d0;
    end if;
  end process p_sync_rst;

  -- Then, instantiate increment counter for VCXO OSC1
  cmp_vcxo20_incr_counter : incr_counter
    generic map
    (
      width => 32
    )
    port map
    (
      clk_i             => clk20_vcxo_i,
      rst_i             => cnt20_actual_rst,
      counter_top_i     => cnt20_top,
      counter_incr_en_i => cnt20_en,
      counter_o         => cnt20,
      counter_is_full_o => cnt20_full
    );

  -- Instantiate clock info slave
  cmp_vcxo20_clk_info : clk_info_wb_slave
    port map
    (
      wb_clk_i          => clk125,
      rst_i             => rst,

      wb_cyc_i          => xbar_master_out(c_slv_clk_info_20).cyc,
      wb_stb_i          => xbar_master_out(c_slv_clk_info_20).stb,
      wb_addr_i         => xbar_master_out(c_slv_clk_info_20).adr(4 downto 2),
      wb_data_i         => xbar_master_out(c_slv_clk_info_20).dat,
      wb_we_i           => xbar_master_out(c_slv_clk_info_20).we,
      wb_data_o         => xbar_master_in(c_slv_clk_info_20).dat,
      wb_ack_o          => xbar_master_in(c_slv_clk_info_20).ack, 

      counter_is_full_i => cnt20_full,
      counter_now_i     => cnt20,
      counter_top_o     => cnt20_top,
      counter_rst_o     => cnt20_rst,
      counter_en_o      => cnt20_en,
      oe_clk_o          => open
    );

  -- Instantate a Wishbone SPI module to control IC18 DAC operation
  cmp_dac_20_spi : wb_spi
    generic map
    (
      g_interface_mode      => CLASSIC,
      g_address_granularity => WORD
    )
    port map
    (
      clk_sys_i  => clk125,
      rst_n_i    => rst_n,

      wb_adr_i   => xbar_master_out(c_slv_dac_spi_20).adr(6 downto 2),
      wb_dat_i   => xbar_master_out(c_slv_dac_spi_20).dat,
      wb_dat_o   => xbar_master_in(c_slv_dac_spi_20).dat,
      wb_sel_i   => xbar_master_out(c_slv_dac_spi_20).sel,
      wb_stb_i   => xbar_master_out(c_slv_dac_spi_20).stb,
      wb_cyc_i   => xbar_master_out(c_slv_dac_spi_20).cyc,
      wb_we_i    => xbar_master_out(c_slv_dac_spi_20).we,
      wb_ack_o   => xbar_master_in(c_slv_dac_spi_20).ack,
      wb_err_o   => xbar_master_in(c_slv_dac_spi_20).err,
      wb_int_o   => xbar_master_in(c_slv_dac_spi_20).int,
      wb_stall_o => xbar_master_in(c_slv_dac_spi_20).stall,

      pad_cs_o   => plldac2_sync_n,
      pad_sclk_o => fpga_plldac2_sclk_o,
      pad_mosi_o => fpga_plldac2_din_o,
      pad_miso_i => '0'
    );
  
  -- Finally, assign the SYNC_N output to the DAC
  fpga_plldac2_sync_n_o <= plldac2_sync_n(0);

  --============================================================================
  -- SFP EEPROM test logic
  -- * test J1 SFP connector using an SFP loopback module
  --============================================================================
  -- First, instantiate an I2C master to handle SFP communication
  cmp_sfp_eeprom_i2c : wb_i2c_master
    generic map
    (
      g_interface_mode      => CLASSIC,
      g_address_granularity => WORD
    )
    port map
    (
      clk_sys_i    => clk125,
      rst_n_i      => rst_n,
      wb_adr_i     => xbar_master_out(c_slv_sfp_i2c).adr(6 downto 2),
      wb_dat_i     => xbar_master_out(c_slv_sfp_i2c).dat,
      wb_we_i      => xbar_master_out(c_slv_sfp_i2c).we,
      wb_stb_i     => xbar_master_out(c_slv_sfp_i2c).stb,
      wb_sel_i     => xbar_master_out(c_slv_sfp_i2c).sel,
      wb_cyc_i     => xbar_master_out(c_slv_sfp_i2c).cyc,
      wb_ack_o     => xbar_master_in(c_slv_sfp_i2c).ack,
      wb_int_o     => xbar_master_in(c_slv_sfp_i2c).int,
      wb_dat_o     => xbar_master_in(c_slv_sfp_i2c).dat,
      scl_pad_i    => i2c_scl_fr_sfp,
      scl_pad_o    => i2c_scl_to_sfp,
      scl_padoen_o => i2c_sfp_scl_en,
      sda_pad_i    => i2c_sda_fr_sfp,
      sda_pad_o    => i2c_sda_to_sfp,
      sda_padoen_o => i2c_sfp_sda_en
    );

  -- and assign the ports and tri-state buffers
  fpga_sfp_mod_def1_b <= i2c_scl_to_sfp when (i2c_sfp_scl_en = '0') else 'Z';
  i2c_scl_fr_sfp      <= fpga_sfp_mod_def1_b;
  fpga_sfp_mod_def2_b <= i2c_sda_to_sfp when (i2c_sfp_sda_en = '0') else 'Z';
  i2c_sda_fr_sfp      <= fpga_sfp_mod_def2_b;
    
  --============================================================================
  -- SFP EEPROM test logic
  -- * test J1 SFP connector using an SFP loopback module
  --============================================================================
  -- First, instantiate an IBUFGDS for MGT clock
  cmp_mgt_clk_ibufds : IBUFDS
    generic map
    (
      DIFF_TERM    => true,
      IBUF_LOW_PWR => false,
      IOSTANDARD   => "DEFAULT"
    )
    port map
    (
      I  => fpga_mgt_clk0_p_i,
      IB => fpga_mgt_clk0_n_i,
      O  => clk_gtp
    );

  -- Connect the MINIC module to the crossbar
  cmp_sfp_minic : xwr_mini_nic
    generic map
    (
      g_interface_mode       => CLASSIC,
      g_address_granularity  => BYTE,
      g_memsize_log2         => c_minic_memsize_log2,
      g_buffer_little_endian => true
    )
    port map
    (
      clk_sys_i           => clk125, 
      rst_n_i             => rst_n,
      mem_data_o          => minic_mem_data_out,
      mem_addr_o          => minic_mem_addr_out,
      mem_data_i          => minic_mem_data_in,
      mem_wr_o            => minic_mem_wr,
      src_o               => minic_src_out,
      src_i               => minic_src_in,
      snk_o               => minic_snk_out,
      snk_i               => minic_snk_in,
      txtsu_port_id_i     => "00000",
      txtsu_frame_id_i    => x"0000",
      txtsu_tsval_i       => x"00000000",
      txtsu_tsincorrect_i => '0',
      txtsu_stb_i         => '0',
      wb_i                => xbar_master_out(c_slv_minic),
      wb_o                => xbar_master_in(c_slv_minic)
    );

  -- Connect the endpoint buffer RAM
  cmp_sfp_dpram : generic_dpram
    generic map 
    (
      g_data_width       => 32,
      g_size             => 2**c_minic_memsize_log2,
      g_with_byte_enable => false,
      g_dual_clock       => false
    )
    port map 
    (
      rst_n_i => rst_n,
      clka_i  => clk125,
      bwea_i  => "0000",
      wea_i   => dpram_we,
      aa_i    => xbar_master_out(c_slv_dpram).adr(c_minic_memsize_log2+1 downto 2),
      da_i    => xbar_master_out(c_slv_dpram).dat,
      qa_o    => xbar_master_in(c_slv_dpram).dat,
      clkb_i  => clk125,
      bweb_i  => "0000",
      web_i   => minic_mem_wr,
      ab_i    => minic_mem_addr_out,
      db_i    => minic_mem_data_out,
      qb_o    => minic_mem_data_in
    );

  -- associate WE to first port of the RAM to WB signals
  dpram_we <= xbar_master_out(c_slv_dpram).cyc and xbar_master_out(c_slv_dpram).stb and
              xbar_master_out(c_slv_dpram).we;

  -- Connect the Endpoint module to the crossbar
  cmp_sfp_endpoint : xwr_endpoint
    generic map 
    (
      g_interface_mode      => CLASSIC,
      g_address_granularity => BYTE,
      g_simulation          => false,
      g_tx_force_gap_length => 0,
      g_pcs_16bit           => false,
      g_rx_buffer_size      => 1024,
      g_with_rx_buffer      => true,
      g_with_flow_control   => false,
      g_with_timestamper    => true,
      g_with_dpi_classifier => false,
      g_with_vlans          => false,
      g_with_rtu            => false,
      g_with_leds           => false,
      g_with_dmtd           => false
    )
    port map 
    (
      clk_ref_i          => clk125,
      clk_sys_i          => clk125,
      clk_dmtd_i         => clk125,
      rst_n_i            => rst_n,
      pps_csync_p1_i     => '0',
      pps_valid_i        => '0',
      phy_rst_o          => phy_rst,
      phy_loopen_o       => phy_loopen,
      phy_ref_clk_i      => clk125,
      phy_tx_data_o      => phy_tx_d,
      phy_tx_k_o         => phy_tx_k,
      phy_tx_disparity_i => phy_tx_disparity,
      phy_tx_enc_err_i   => phy_tx_error,
      phy_rx_data_i      => phy_rx_d,
      phy_rx_clk_i       => phy_rx_clk,
      phy_rx_k_i         => phy_rx_k,
      phy_rx_enc_err_i   => phy_rx_error,
      phy_rx_bitslide_i  => phy_rx_bitslide,

      src_o => minic_snk_in,
      src_i => minic_snk_out,
      snk_o => minic_src_in,
      snk_i => minic_src_out,

      wb_i => xbar_master_out(c_slv_endpoint),
      wb_o => xbar_master_in(c_slv_endpoint)
    );

  -- Finally, connect the GTP transceiver to the endpoint signals
  cmp_gtp_xceiver : wr_gtp_phy_spartan6
    generic map 
    (
      g_simulation      => 0,
      g_force_disparity => 1,
      g_enable_ch0      => 1,
      g_enable_ch1      => 0
    )
    port map 
    (
      gtp_clk_i          => clk_gtp,
      ch0_ref_clk_i      => clk125,
      ch0_tx_data_i      => phy_tx_d(7 downto 0),
      ch0_tx_k_i         => phy_tx_k(0),
      ch0_tx_disparity_o => phy_tx_disparity,
      ch0_tx_enc_err_o   => phy_tx_error,
      ch0_rx_rbclk_o     => phy_rx_clk,
      ch0_rx_data_o      => phy_rx_d(7 downto 0),
      ch0_rx_k_o         => phy_rx_k(0),
      ch0_rx_enc_err_o   => phy_rx_error,
      ch0_rx_bitslide_o  => phy_rx_bitslide(3 downto 0),
      ch0_rst_i          => phy_rst,
      ch0_loopen_i       => phy_loopen,
      pad_txn0_o         => mgt_sfp_tx0_n_o,
      pad_txp0_o         => mgt_sfp_tx0_p_o,
      pad_rxn0_i         => mgt_sfp_rx0_n_i,
      pad_rxp0_i         => mgt_sfp_rx0_p_i
    );

  --============================================================================
  -- LED test logic
  -- * test bicolor LEDs and its driving circuit (IC1)
  -- * test front panel LED logic and driving circuit (IC5)
  --============================================================================
  -- Process to control the LED sequence counter
  p_led_seq : process (clk125) is
  begin
    if rising_edge(clk125) then
      if (rst_n = '0') then
        cnt_halfsec <= (others => '0');
        led_seq     <= (others => '0');
      elsif (pulse_led_en = '1') or (stat_led_en = '1') then
        cnt_halfsec <= cnt_halfsec + 1;
        if (cnt_halfsec = 62499999) then
          cnt_halfsec <= (others => '0');
          led_seq     <= led_seq + 1;
          if (led_seq = 24) then
            led_seq   <= (others => '0');
          end if;
        end if;
      end if;
    end if;
  end process p_led_seq;

  -- Sequence the front-panel LEDs based on the sequence counter
  pulse_front_led_n_o <= "111110" when (pulse_led_en = '1') and (led_seq =  1) else
                         "111101" when (pulse_led_en = '1') and (led_seq =  2) else 
                         "111011" when (pulse_led_en = '1') and (led_seq =  3) else 
                         "110111" when (pulse_led_en = '1') and (led_seq =  4) else 
                         "101111" when (pulse_led_en = '1') and (led_seq =  5) else 
                         "011111" when (pulse_led_en = '1') and (led_seq =  6) else 
                         "111110" when (pulse_led_en = '1') and (led_seq =  7) else
                         "111101" when (pulse_led_en = '1') and (led_seq =  8) else 
                         "111011" when (pulse_led_en = '1') and (led_seq =  9) else 
                         "110111" when (pulse_led_en = '1') and (led_seq = 10) else 
                         "101111" when (pulse_led_en = '1') and (led_seq = 11) else 
                         "011111" when (pulse_led_en = '1') and (led_seq = 12) else 
                         "111110" when (pulse_led_en = '1') and (led_seq = 13) else
                         "111101" when (pulse_led_en = '1') and (led_seq = 14) else 
                         "111011" when (pulse_led_en = '1') and (led_seq = 15) else 
                         "110111" when (pulse_led_en = '1') and (led_seq = 16) else 
                         "101111" when (pulse_led_en = '1') and (led_seq = 17) else 
                         "011111" when (pulse_led_en = '1') and (led_seq = 18) else 
                         "111110" when (pulse_led_en = '1') and (led_seq = 19) else
                         "111101" when (pulse_led_en = '1') and (led_seq = 20) else 
                         "111011" when (pulse_led_en = '1') and (led_seq = 21) else 
                         "110111" when (pulse_led_en = '1') and (led_seq = 22) else 
                         "101111" when (pulse_led_en = '1') and (led_seq = 23) else 
                         "011111" when (pulse_led_en = '1') and (led_seq = 24) else 
                         "111111";

  -- Light each LED red and green in a sequence, based on the sequence counter.
  --
  -- The colors are set via the LED state vector (two bits per LED) as follows:
  --    state(1..0)     color
  --        00           OFF
  --        01           RED
  --        10          GREEN
  bicolor_led_state <= "000000000001000000000000" when (stat_led_en = '1') and (led_seq =  1) else
                       "000000000100000000000000" when (stat_led_en = '1') and (led_seq =  2) else
                       "000000010000000000000000" when (stat_led_en = '1') and (led_seq =  3) else
                       "000001000000000000000000" when (stat_led_en = '1') and (led_seq =  4) else
                       "000000000000000001000000" when (stat_led_en = '1') and (led_seq =  5) else
                       "000000000000000000010000" when (stat_led_en = '1') and (led_seq =  6) else
                       "000000000000000000000100" when (stat_led_en = '1') and (led_seq =  7) else
                       "000000000000000000000001" when (stat_led_en = '1') and (led_seq =  8) else
                       "000000000000000100000000" when (stat_led_en = '1') and (led_seq =  9) else
                       "000000000000010000000000" when (stat_led_en = '1') and (led_seq = 10) else
                       "000100000000000000000000" when (stat_led_en = '1') and (led_seq = 11) else
                       "010000000000000000000000" when (stat_led_en = '1') and (led_seq = 12) else
                       "000000000010000000000000" when (stat_led_en = '1') and (led_seq = 13) else
                       "000000001000000000000000" when (stat_led_en = '1') and (led_seq = 14) else
                       "000000100000000000000000" when (stat_led_en = '1') and (led_seq = 15) else
                       "000010000000000000000000" when (stat_led_en = '1') and (led_seq = 16) else
                       "000000000000000010000000" when (stat_led_en = '1') and (led_seq = 17) else
                       "000000000000000000100000" when (stat_led_en = '1') and (led_seq = 18) else
                       "000000000000000000001000" when (stat_led_en = '1') and (led_seq = 19) else
                       "000000000000000000000010" when (stat_led_en = '1') and (led_seq = 20) else
                       "000000000000001000000000" when (stat_led_en = '1') and (led_seq = 21) else
                       "000000000000100000000000" when (stat_led_en = '1') and (led_seq = 22) else
                       "001000000000000000000000" when (stat_led_en = '1') and (led_seq = 23) else
                       "100000000000000000000000" when (stat_led_en = '1') and (led_seq = 24) else
                       "000000000000000000000000";

  -- Then, we instantiate the LED controller and control it via the LED state
  -- vector.
  cmp_bicolor_led_ctrl : bicolor_led_ctrl
    generic map
    (
      g_NB_COLUMN    => 6,
      g_NB_LINE      => 2,
      g_clk_freq     => 125000000,
      g_refresh_rate => 250
    )
    port map
    (
      clk_i           => clk125,
      rst_n_i         => rst_n,
      led_intensity_i => "1111111",
      led_state_i     => bicolor_led_state,
      column_o(0)     => led_wr_ownaddr_i2c_o,
      column_o(1)     => led_wr_gmt_ttl_ttln_o,
      column_o(2)     => led_wr_link_syserror_o,
      column_o(3)     => led_wr_ok_syspw_o,
      column_o(4)     => led_multicast_2_0_o,
      column_o(5)     => led_multicast_3_1_o,
      line_o(0)       => led_ctrl0_o,
      line_o(1)       => led_ctrl1_o,
      line_oen_o(0)   => led_ctrl0_oen_o,
      line_oen_o(1)   => led_ctrl1_oen_o
    );
  
end behav;
