--------------------------------------------------------------------------------
-- CERN (BE-CO-HT)
-- Top level entity of CONV-TTL-BLO V1
-- http://www.ohwr.org/projects/conv-trig-blo
--------------------------------------------------------------------------------
--
-- unit name: conv_ttl_blo_v2.vhd
--
-- author: Theodor-Adrian Stana (t.stana@cern.ch)
--         Carlos Gil Soriano (gilsoriano@gmail.com)
--
-- version: 1.0
--
-- description: Top entity of CONV-TTL-BLO V1
--
-- dependencies:
--
--------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
--------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--------------------------------------------------------------------------------
library IEEE;
library unisim;

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use UNISIM.VCOMPONENTS.ALL;

use work.bicolor_led_ctrl_pkg.all;

entity conv_ttl_blo_v2 is
  generic
  (
    g_nr_ttl_chan : natural := 6;
    g_nr_inv_chan : natural := 4
  );
  port 
  (
--    RST                  : in  std_logic;
--    CLK20_VCXO           : in  std_logic;
    FPGA_CLK_P           : in  std_logic; --Using the 125MHz clock
    FPGA_CLK_N           : in  std_logic;
    
    -- LEDs
    LED_CTRL0            : out std_logic;
    LED_CTRL0_OEN        : out std_logic;
    LED_CTRL1            : out std_logic;
    LED_CTRL1_OEN        : out std_logic;
    LED_MULTICAST_2_0    : out std_logic;
    LED_MULTICAST_3_1    : out std_logic;
    LED_WR_GMT_TTL_TTLN  : out std_logic;
    LED_WR_LINK_SYSERROR : out std_logic;
    LED_WR_OK_SYSPW      : out std_logic;
    LED_WR_OWNADDR_I2C   : out std_logic;
    
    -- I/Os for pulses
    PULSE_FRONT_LED_N    : out std_logic_vector(g_nr_ttl_chan downto 1);
    PULSE_REAR_LED_N     : out std_logic_vector(g_nr_ttl_chan downto 1);
    FPGA_INPUT_TTL_N     : in  std_logic_vector(g_nr_ttl_chan downto 1);
    FPGA_OUT_TTL         : out std_logic_vector(g_nr_ttl_chan downto 1);
    FPGA_BLO_IN          : in  std_logic_vector(g_nr_ttl_chan downto 1);
    FPGA_TRIG_BLO        : out std_logic_vector(g_nr_ttl_chan downto 1);
    INV_IN_N             : in  std_logic_vector(g_nr_inv_chan downto 1);
    INV_OUT              : out std_logic_vector(g_nr_inv_chan downto 1);
    
--    -- Lines for the i2c_slave
--    SCL_I                : in  std_logic;
--    SCL_O                : out std_logic;
--    SCL_OE               : out std_logic;
--    SDA_I                : in  std_logic;
--    SDA_O                : out std_logic;
--    SDA_OE               : out std_logic;
--    FPGA_GA              : in  std_logic_vector(4 downto 0);
--    FPGA_GAP             : in  std_logic;
--    
--    -- Pins of the SPI interface to write into the Flash memory
--    FPGA_PROM_CCLK       : out std_logic;
--    FPGA_PROM_CSO_B_N    : out std_logic;
--    FPGA_PROM_DIN        : in  std_logic;
--    FPGA_PROM_MOSI       : out std_logic;
--      
    FPGA_OE              : out std_logic;
    FPGA_BLO_OE          : out std_logic;
    FPGA_TRIG_TTL_OE     : out std_logic;
    FPGA_INV_OE          : out std_logic;
    
    --TTL/INV_TTL_N
    LEVEL                : in  std_logic;
    
    EXTRA_SWITCH         : in  std_logic_vector(7 downto 1);
    
    -- It allows power sequencing of the
    -- 24V rail after a security given delay
    MR_N                 : out std_logic;

    -- RTM identifiers, should match with the expected values
    -- TODO: add matching
    FPGA_RTMM_N        : in std_logic_vector(2 downto 0);
    FPGA_RTMP_N        : in std_logic_vector(2 downto 0)
  );
end conv_ttl_blo_v2;


architecture behav of conv_ttl_blo_v2 is


  --============================================================================
  -- Component declarations
  --============================================================================
  -- Reset generator component
  -- (use: global reset generation, output reset generation)
  component reset_gen is
    generic
    (
      -- Reset time in number of clk_i cycles
      g_reset_time : positive := 5_000_000
    );
    port
    (
      clk_i   : in  std_logic;
      rst_i   : in  std_logic;
      rst_n_o : out std_logic
    );
  end component reset_gen;

  -- Pulse generator component
  -- (use: output pulse generation, pulse status LEDs)
  component pulse_generator is
    generic 
    (
      -- Pulse width, in number of clk_i cycles
      g_pulse_width     : natural := 15;

      -- Glitch filter length:
      -- g_glitch_filt_len=1 => trigger width should be > 1 clk_i cycle
      -- g_glitch_filt_len=2 => trigger width should be > 2 clk_i cycles
      -- etc.
      g_glitch_filt_len : natural := 4

    );
    port
    (
      clk_i        : in  std_logic;
      rst_n_i      : in  std_logic;

      -- Pulse type input
      -- '1' - TYPE 1 (glitch-sensitive, without output jitter)
      -- '0' - TYPE 2 (glitch-insensitive, with output jitter)
      pulse_type_i : in  std_logic;

      -- Enable input, pulse generation is enabled when '1'
      enable_i     : in  std_logic;

      -- Trigger input, has to be '1' to assure pulse output with delay no greater
      -- than internal gate delays.
      trig_i       : in  std_logic;

      -- Pulse output, active-high
      -- latency:
      --    TYPE 1 pulse: none
      --    TYPE 2 pulse: g_glitch_filt_len+3 clk_i cycles
      pulse_o      : out std_logic
    );
  end component pulse_generator;

  -- RTM detector component
  -- (use: detect the presence of an RTM/P module)
  component rtm_detector is
    port 
    (
      rtmm_i    : in std_logic_vector(2 downto 0);
      rtmp_i    : in std_logic_vector(2 downto 0);
      rtmm_ok_o : out std_logic;
      rtmp_ok_o : out std_logic
    );
  end component rtm_detector;

  component pulse_gen is
    generic
    (
      g_pwidth : natural := 200;
      g_freq   : natural := 400
    );
    port
    (
      clk_i   : in  std_logic;
      rst_n_i : in  std_logic;
      pulse_o : out std_logic
    );
  end component pulse_gen;
  
  --============================================================================
  -- Signal declarations
  --============================================================================
  -- Clock signals
  signal clk_125                    : std_logic;
  signal clk_50, clk_buf_50         : std_logic;
  signal clk_200, clk_buf_200       : std_logic;
  signal clk_250, clk_buf_250       : std_logic;
  signal pll_fb_in, pll_fb_out      : std_logic;
  signal pll_locked                 : std_logic;

  -- Reset signals
  signal rst_n, rst                 : std_logic;

  -- RTM detection signals
  signal rtmm, rtmp                 : std_logic_vector(2 downto 0);
  signal rtmm_ok, rtmp_ok           : std_logic;

  -- Signals for pulse generation triggers
  signal trig                       : std_logic_vector(g_nr_ttl_chan downto 1);
  signal trig_inv                   : std_logic_vector(g_nr_inv_chan downto 1);
  signal trig_ttl, trig_blo         : std_logic_vector(g_nr_ttl_chan downto 1);

  -- Temporary signal for blocking and TTL pulse outputs
  signal pulse_outputs              : std_logic_vector(g_nr_ttl_chan downto 1);
  
  -- Temporary signal for inverted-TTL pulse outputs
  signal inv_outputs                : std_logic_vector(g_nr_inv_chan downto 1);
  
  -- Pulse status LED signals
  signal front_led_en               : std_logic_vector(g_nr_ttl_chan downto 1);
  signal rear_led_en                : std_logic_vector(g_nr_ttl_chan downto 1);
  signal pulse_leds                 : std_logic_vector(g_nr_ttl_chan downto 1);

  -- Output enable signals
  signal oe, ttl_oe, blo_oe, inv_oe : std_logic;

  -- Signal for controlling the bicolor LED matrix
  signal bicolor_led_state          : std_logic_vector(23 downto 0);

  signal tmp_pulse : std_logic;
  
  
begin

  --============================================================================
  -- Generate 125 MHz global signal from differential lines
  --============================================================================
  cmp_125_diff_buf: IBUFGDS
  generic map 
  (
    DIFF_TERM    => TRUE,
    IBUF_LOW_PWR => TRUE
  )
  port map 
  (
    I  => FPGA_CLK_P,
    IB => FPGA_CLK_N,
    O  => clk_125
  );

  --============================================================================
  -- Internal and external reset generation
  --============================================================================
  -- Configure reset generator for 96ms power-on reset
  cmp_reset_gen: reset_gen
    generic map
    (
      -- Reset time: 12 * 8ns * (10**6) = 96 ms
      g_reset_time => 12*(10**6)
    )
    port map
    (
      clk_i   => clk_125,
      rst_i   => '0',
      rst_n_o => rst_n
    );
  
--  rst  <= not rst_n;
  MR_N <= rst_n;

  --============================================================================
  -- Output enable logic
  --============================================================================
  -- The general output enable is set first and the blocking, TTL
  -- and INV output enable signals are set one clock cycle later.
  p_oe: process(clk_125)
  begin
    if rising_edge(clk_125) then
      if (rst_n = '0') then
        oe     <= '0';
        blo_oe <= '0';
        ttl_oe <= '0';
        inv_oe <= '0';
      else
        oe <= '1';
        if (oe = '1') then
          blo_oe <= '1';
          ttl_oe <= '1';
          inv_oe <= '1';
        end if;
      end if;
    end if;
  end process p_oe;

  FPGA_OE          <= oe;
  FPGA_BLO_OE      <= blo_oe;
  FPGA_TRIG_TTL_OE <= ttl_oe;
  FPGA_INV_OE      <= inv_oe;

  --============================================================================
  -- TTL and blocking pulse generation logic
  --============================================================================
  trig_ttl <= not FPGA_INPUT_TTL_N when (LEVEL = '0') else 
              FPGA_INPUT_TTL_N;
  trig_blo <= FPGA_BLO_IN;
  trig     <= trig_ttl or trig_blo;

  gen_ttl_pulse_generators: for i in 1 to g_nr_ttl_chan generate
    -- Output pulse generators
    cmp_ttl_pulse_gen: pulse_generator
      generic map
      (
        g_pulse_width     => 125,
        g_glitch_filt_len => 4
      )
      port map
      (
        clk_i        => clk_125,
        rst_n_i      => rst_n,
        pulse_type_i => EXTRA_SWITCH(1),
        enable_i     => '1',
        trig_i       => trig(i),
        pulse_o      => pulse_outputs(i)
      );

    -- Status LED pulse generators
    cmp_led_pulse_gen: pulse_generator
      generic map
      (
        g_pulse_width     => 12*(10**6),
        g_glitch_filt_len => 4
      )
      port map
      (
        clk_i        => clk_125,
        rst_n_i      => rst_n,
        enable_i     => '1',
        pulse_type_i => EXTRA_SWITCH(1),
        trig_i       => trig(i),
        pulse_o      => pulse_leds(i)
      );
  end generate gen_ttl_pulse_generators;
    
  -- Pulse outputs assignment
  FPGA_OUT_TTL  <= pulse_outputs;
  FPGA_TRIG_BLO <= pulse_outputs;
  
  -- Pulse status LED output assignments
  PULSE_FRONT_LED_N <= (not pulse_leds) when (ttl_oe = '1') else 
                       (others => '0');
  PULSE_REAR_LED_N  <= (not pulse_leds) when (blo_oe = '1') else 
                       (others => '0');

  --============================================================================
  -- Inverted TTL pulse generation logic
  --============================================================================
  -- Trigger for pulse generators
  trig_inv <= not INV_IN_N when (LEVEL = '0') else
              INV_IN_N;
  
  -- Instantiate the necessary number of pulse generator components  
  gen_inv_pulse_generators: for i in 1 to g_nr_inv_chan generate
    cmp_inv_pulse_gen: pulse_generator
      generic map
      (
        g_pulse_width     => 125,
        g_glitch_filt_len => 4
      )
      port map
      (
        clk_i        => clk_125,
        rst_n_i      => rst_n,
        pulse_type_i => EXTRA_SWITCH(1),
        enable_i     => '1',
        trig_i       => trig_inv(i),
        pulse_o      => inv_outputs(i)
      );
  end generate gen_inv_pulse_generators;

  -- Output assignment
  INV_OUT <= not inv_outputs;
  
--  cmp_tmp_pulse_gen: pulse_gen
--    generic map
--    (
--      g_pwidth => 100,
--      g_freq   => 125*(10**6)
--    )
--    port map
--    (
--      clk_i   => clk_125,
--      rst_n_i => rst_n,
--      pulse_o => tmp_pulse
--    );
  
  --============================================================================
  -- Bicolor LED matrix logic
  --============================================================================
  -- Bicolor LED controls, corresponding to the column orders on the 
  -- bicolor_led_ctrl unit.

  -- WR address
  bicolor_led_state( 1 downto  0) <= c_LED_OFF;

  -- WR GMT
  bicolor_led_state( 3 downto  2) <= c_LED_OFF;

  -- WR link
  bicolor_led_state( 5 downto  4) <= c_LED_OFF;

  -- WR OK
  bicolor_led_state( 7 downto  6) <= c_LED_OFF;

  -- MULTICAST 0
  bicolor_led_state( 9 downto  8) <= c_LED_OFF;

  -- MULTICAST 1
  bicolor_led_state(11 downto 10) <= c_LED_OFF;


  -- I2C valid
  bicolor_led_state(13 downto 12) <= c_LED_OFF;

  -- State of TTL/TTL_N switch
  bicolor_led_state(15 downto 14) <= c_LED_GREEN when (LEVEL = '1') else
                                     c_LED_OFF;

  -- System error
  bicolor_led_state(17 downto 16) <= c_LED_RED when (rtmm_ok = '0') and (rtmp_ok = '0') else
                                     c_LED_OFF;

  -- System power
  bicolor_led_state(19 downto 18) <= c_LED_GREEN;

  -- MULTICAST 2
  bicolor_led_state(21 downto 20) <= c_LED_OFF; 

  -- MULTICAST 3
  bicolor_led_state(23 downto 22) <= c_LED_OFF;

  cmp_bicolor_led_ctrl: bicolor_led_ctrl
    generic map
    (
      g_NB_COLUMN    => 6,
      g_NB_LINE      => 2,
      g_clk_freq     => 125000000,
      g_refresh_rate => 250
    )
    port map
    (
      clk_i           => clk_125,
      rst_n_i         => rst_n,
      led_intensity_i => "0011111",
      led_state_i     => bicolor_led_state,
      column_o(0)     => LED_WR_OWNADDR_I2C,
      column_o(1)     => LED_WR_GMT_TTL_TTLN,
      column_o(2)     => LED_WR_LINK_SYSERROR,
      column_o(3)     => LED_WR_OK_SYSPW,
      column_o(4)     => LED_MULTICAST_2_0,
      column_o(5)     => LED_MULTICAST_3_1,
      line_o(0)       => LED_CTRL0,
      line_o(1)       => LED_CTRL1,
      line_oen_o(0)   => LED_CTRL0_OEN,
      line_oen_o(1)   => LED_CTRL1_OEN
    );

  --============================================================================
  -- RTM detection logic
  --============================================================================
  rtmm <= not FPGA_RTMM_N;
  rtmp <= not FPGA_RTMP_N;

  cmp_rtm_detector: rtm_detector
    port map
    (
      rtmm_i    => rtmm,
      rtmp_i    => rtmp,
      rtmm_ok_o => rtmm_ok,
      rtmp_ok_o => rtmp_ok
    );

end behav;
